# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import scipy.stats as st
import numpy as np
import pandas as pd
import time, os, logging

from allensdk.brain_observatory.findlevel import findlevel
from allensdk.brain_observatory.brain_observatory_exceptions import \
    BrainObservatoryAnalysisException

class StimulusAnalysis(object):
    """ Base class for all response analysis code. Subclasses are responsible
    for computing metrics and traces relevant to a particular stimulus.  
    The base class contains methods for organizing sweep responses row of 
    a stimulus stable (get_sweep_response).  Subclasses implement the
    get_response method, computes the mean sweep response to all sweeps for
    a each stimulus condition.    

    Parameters
    ----------
    data_set: BrainObservatoryNwbDataSet instance

    speed_tuning: boolean
       Whether or not to compute speed tuning histograms

    """
    _log = logging.getLogger('allensdk.brain_observatory.stimulus_analysis')
    
    def __init__(self, data_set, speed_tuning=False):
        self.data_set = data_set
        
        # get fluorescence 
        self.timestamps, self.celltraces = self.data_set.get_corrected_fluorescence_traces()
        self.numbercells = len(self.celltraces)    #number of cells in dataset
        self.roi_id = self.data_set.get_roi_ids()
        self.cell_id = self.data_set.get_cell_specimen_ids()
        
        # get dF/F 
        _, self.dfftraces = self.data_set.get_dff_traces()

        self.acquisition_rate = 1/(self.timestamps[1]-self.timestamps[0])
        self.dxcm, self.dxtime = self.data_set.get_running_speed()

        if speed_tuning:
            self.binned_dx_sp, self.binned_cells_sp, self.binned_dx_vis, self.binned_cells_vis, self.peak_run = self.get_speed_tuning(binsize=800)

    def get_response(self):
        """ Implemented by subclasses. """
        raise BrainObservatoryAnalysisException("get_response not implemented")

    def get_peak(self):
        """ Implemented by subclasses. """
        raise BrainObservatoryAnalysisException("get_peak not implemented")

    def get_speed_tuning(self, binsize):
        """ Calculates speed tuning, spontaneous versus visually driven.  The return is a 5-tuple 
        of speed and dF/F histograms.  

            binned_dx_sp: (bins,2) np.ndarray of running speeds binned during spontaneous activity stimulus.  
            The first bin contains all speeds below 1 cm/s.  Dimension 0 is mean running speed in the bin.
            Dimension 1 is the standard error of the mean.

            binned_cells_sp: (bins,2) np.ndarray of fluorescence during spontaneous activity stimulus.  
            First bin contains all data for speeds below 1 cm/s. Dimension 0 is mean fluorescence in the bin.
            Dimension 1 is the standard error of the mean.

            binned_dx_vis: (bins,2) np.ndarray of running speeds outside of spontaneous activity stimulus.
            The first bin contains all speeds below 1 cm/s.  Dimension 0 is mean running speed in the bin.
            Dimension 1 is the standard error of the mean.

            binned_cells_vis: np.ndarray of fluorescence outside of spontaneous activity stimulu.
            First bin contains all data for speeds below 1 cm/s. Dimension 0 is mean fluorescence in the bin.
            Dimension 1 is the standard error of the mean.

            peak_run: pd.DataFrame of speed-related properties of a cell.
            
        Returns
        -------
        tuple: binned_dx_sp, binned_cells_sp, binned_dx_vis, binned_cells_vis, peak_run
        """

        StimulusAnalysis._log.info('Calculating speed tuning, spontaneous vs visually driven')
        
        celltraces_trimmed = np.delete(self.dfftraces, range(len(self.dxcm), np.size(self.dfftraces,1)), axis=1) 

        # pull out spontaneous epoch(s)        
        spontaneous = self.data_set.get_stimulus_table('spontaneous')

        peak_run = pd.DataFrame(index=range(self.numbercells), columns=('speed_max_sp','speed_min_sp','ptest_sp', 'mod_sp','speed_max_vis','speed_min_vis','ptest_vis', 'mod_vis'))

        dx_sp = self.dxcm[spontaneous.start.iloc[-1]:spontaneous.end.iloc[-1]]
        celltraces_sp = celltraces_trimmed[:,spontaneous.start.iloc[-1]:spontaneous.end.iloc[-1]]
        dx_vis = np.delete(self.dxcm, np.arange(spontaneous.start.iloc[-1],spontaneous.end.iloc[-1]))
        celltraces_vis = np.delete(celltraces_trimmed, np.arange(spontaneous.start.iloc[-1],spontaneous.end.iloc[-1]), axis=1)
        if len(spontaneous) > 1:
            dx_sp = np.append(dx_sp, self.dxcm[spontaneous.start.iloc[-2]:spontaneous.end.iloc[-2]], axis=0)
            celltraces_sp = np.append(celltraces_sp,celltraces_trimmed[:,spontaneous.start.iloc[-2]:spontaneous.end.iloc[-2]], axis=1)
            dx_vis = np.delete(dx_vis, np.arange(spontaneous.start.iloc[-2],spontaneous.end.iloc[-2]))
            celltraces_vis = np.delete(celltraces_vis, np.arange(spontaneous.start.iloc[-2],spontaneous.end.iloc[-2]), axis=1)
        celltraces_vis = celltraces_vis[:,~np.isnan(dx_vis)]
        dx_vis = dx_vis[~np.isnan(dx_vis)]  

        nbins = 1 + len(np.where(dx_sp>=1)[0])/binsize
        dx_sorted = dx_sp[np.argsort(dx_sp)]
        celltraces_sorted_sp = celltraces_sp[:, np.argsort(dx_sp)]
        binned_cells_sp = np.zeros((self.numbercells, nbins, 2))
        binned_dx_sp = np.zeros((nbins,2))
        for i in range(nbins):
            if np.all(np.isnan(dx_sorted)):
                raise BrainObservatoryAnalysisException("dx is filled with NaNs")

            offset = findlevel(dx_sorted,1,'up')

            if offset is None:
                raise BrainObservatoryAnalysisException("Could not find crossing in dx")

            if i==0:
                binned_dx_sp[i,0] = np.mean(dx_sorted[:offset])
                binned_dx_sp[i,1] = np.std(dx_sorted[:offset]) / np.sqrt(offset)            
                binned_cells_sp[:,i,0] = np.mean(celltraces_sorted_sp[:,:offset], axis=1)
                binned_cells_sp[:,i,1] = np.std(celltraces_sorted_sp[:,:offset], axis=1) / np.sqrt(offset)
            else:
                start = offset + (i-1)*binsize
                binned_dx_sp[i,0] = np.mean(dx_sorted[start:start+binsize])
                binned_dx_sp[i,1] = np.std(dx_sorted[start:start+binsize])/np.sqrt(binsize)
                binned_cells_sp[:,i,0] = np.mean(celltraces_sorted_sp[:, start:start+binsize], axis=1)
                binned_cells_sp[:,i,1] = np.std(celltraces_sorted_sp[:,start:start+binsize], axis=1)/np.sqrt(binsize)
        
        binned_cells_shuffled_sp = np.empty((self.numbercells, nbins, 2, 200))
        for shuf in range(200):
            celltraces_shuffled = celltraces_sp[:,np.random.permutation(np.size(celltraces_sp,1))]
            celltraces_shuffled_sorted = celltraces_shuffled[:, np.argsort(dx_sp)]
            for i in range(nbins):
                offset = findlevel(dx_sorted,1,'up')        
                if i==0:          
                    binned_cells_shuffled_sp[:,i,0,shuf] = np.mean(celltraces_shuffled_sorted[:,:offset], axis=1)
                    binned_cells_shuffled_sp[:,i,1,shuf] = np.std(celltraces_shuffled_sorted[:,:offset], axis=1)
                else:
                    start = offset + (i-1)*binsize
                    binned_cells_shuffled_sp[:,i,0,shuf] = np.mean(celltraces_shuffled_sorted[:, start:start+binsize], axis=1)
                    binned_cells_shuffled_sp[:,i,1,shuf] = np.std(celltraces_shuffled_sorted[:,start:start+binsize], axis=1)
                
        
        nbins = 1 + len(np.where(dx_vis>=1)[0])/binsize
        dx_sorted = dx_vis[np.argsort(dx_vis)]
        celltraces_sorted_vis = celltraces_vis[:, np.argsort(dx_vis)]
        binned_cells_vis = np.zeros((self.numbercells, nbins, 2))
        binned_dx_vis = np.zeros((nbins,2))
        for i in range(nbins):
            offset = findlevel(dx_sorted,1,'up')        
            if i==0:
                binned_dx_vis[i,0] = np.mean(dx_sorted[:offset])
                binned_dx_vis[i,1] = np.std(dx_sorted[:offset]) / np.sqrt(offset)
                binned_cells_vis[:,i,0] = np.mean(celltraces_sorted_vis[:,:offset], axis=1)
                binned_cells_vis[:,i,1] = np.std(celltraces_sorted_vis[:,:offset], axis=1) / np.sqrt(offset)
            else:
                start = offset + (i-1)*binsize
                binned_dx_vis[i,0] = np.mean(dx_sorted[start:start+binsize])
                binned_dx_vis[i,1] = np.std(dx_sorted[start:start+binsize])/np.sqrt(binsize)
                binned_cells_vis[:,i,0] = np.mean(celltraces_sorted_vis[:, start:start+binsize], axis=1)
                binned_cells_vis[:,i,1] = np.std(celltraces_sorted_vis[:,start:start+binsize], axis=1)/np.sqrt(binsize)
        
        binned_cells_shuffled_vis = np.empty((self.numbercells, nbins, 2, 200))
        for shuf in range(200):
            celltraces_shuffled = celltraces_vis[:,np.random.permutation(np.size(celltraces_vis,1))]
            celltraces_shuffled_sorted = celltraces_shuffled[:, np.argsort(dx_vis)]
            for i in range(nbins):
                offset = findlevel(dx_sorted,1,'up')        
                if i==0:          
                    binned_cells_shuffled_vis[:,i,0,shuf] = np.mean(celltraces_shuffled_sorted[:,:offset], axis=1)
                    binned_cells_shuffled_vis[:,i,1,shuf] = np.std(celltraces_shuffled_sorted[:,:offset], axis=1)
                else:
                    start = offset + (i-1)*binsize
                    binned_cells_shuffled_vis[:,i,0,shuf] = np.mean(celltraces_shuffled_sorted[:, start:start+binsize], axis=1)
                    binned_cells_shuffled_vis[:,i,1,shuf] = np.std(celltraces_shuffled_sorted[:,start:start+binsize], axis=1)
         
        shuffled_variance_sp = binned_cells_shuffled_sp[:,:,0,:].std(axis=1)**2
        variance_threshold_sp = np.percentile(shuffled_variance_sp, 99.9, axis=1)
        response_variance_sp = binned_cells_sp[:,:,0].std(axis=1)**2

                 
        shuffled_variance_vis = binned_cells_shuffled_vis[:,:,0,:].std(axis=1)**2
        variance_threshold_vis = np.percentile(shuffled_variance_vis, 99.9, axis=1)
        response_variance_vis = binned_cells_vis[:,:,0].std(axis=1)**2
         
        for nc in range(self.numbercells):
            if response_variance_vis[nc]>variance_threshold_vis[nc]:
                peak_run.mod_vis[nc] = True
            if response_variance_vis[nc]<=variance_threshold_vis[nc]:
                peak_run.mod_vis[nc] = False
            if response_variance_sp[nc]>variance_threshold_sp[nc]:
                peak_run.mod_sp[nc] = True
            if response_variance_sp[nc]<=variance_threshold_sp[nc]:
                peak_run.mod_sp[nc] = False
            temp = binned_cells_sp[nc,:,0]
            start_max = temp.argmax()
            peak_run.speed_max_sp[nc] = binned_dx_sp[start_max,0]
            start_min = temp.argmin()
            peak_run.speed_min_sp[nc] = binned_dx_sp[start_min,0]
            if peak_run.speed_max_sp[nc]>peak_run.speed_min_sp[nc]:
                test_values = celltraces_sorted_sp[nc,start_max*binsize:(start_max+1)*binsize]
                other_values = np.delete(celltraces_sorted_sp[nc,:], range(start_max*binsize, (start_max+1)*binsize))
                (_ ,peak_run.ptest_sp[nc]) = st.ks_2samp(test_values, other_values)
            else:
                test_values = celltraces_sorted_sp[nc,start_min*binsize:(start_min+1)*binsize]
                other_values = np.delete(celltraces_sorted_sp[nc,:], range(start_min*binsize, (start_min+1)*binsize))
                (_ ,peak_run.ptest_sp[nc]) = st.ks_2samp(test_values, other_values)
            temp = binned_cells_vis[nc,:,0]
            start_max = temp.argmax()
            peak_run.speed_max_vis[nc] = binned_dx_vis[start_max,0]
            start_min = temp.argmin()
            peak_run.speed_min_vis[nc] = binned_dx_vis[start_min,0]
            if peak_run.speed_max_vis[nc]>peak_run.speed_min_vis[nc]:
                test_values = celltraces_sorted_vis[nc,start_max*binsize:(start_max+1)*binsize]
                other_values = np.delete(celltraces_sorted_vis[nc,:], range(start_max*binsize, (start_max+1)*binsize))
            else:  
                test_values = celltraces_sorted_vis[nc,start_min*binsize:(start_min+1)*binsize]
                other_values = np.delete(celltraces_sorted_vis[nc,:], range(start_min*binsize, (start_min+1)*binsize))
            (_ ,peak_run.ptest_vis[nc]) = st.ks_2samp(test_values, other_values)
        
        return binned_dx_sp, binned_cells_sp, binned_dx_vis, binned_cells_vis, peak_run


    def get_sweep_response(self):
        """ Calculates the response to each sweep in the stimulus table for each cell and the mean response.
        The return is a 3-tuple of:

            * sweep_response: pd.DataFrame of response dF/F traces organized by cell (column) and sweep (row)

            * mean_sweep_response: mean values of the traces returned in sweep_response

            * pval: p value from 1-way ANOVA comparing response during sweep to response prior to sweep

        Returns
        -------
        3-tuple: sweep_response, mean_sweep_response, pval
        """
        def do_mean(x):
            return np.mean(x[self.interlength:self.interlength+self.sweeplength+self.extralength])#+1])
            
        def do_p_value(x):
            (_, p) = st.f_oneway(x[:self.interlength], x[self.interlength:self.interlength+self.sweeplength+self.extralength])
            return p
            
        StimulusAnalysis._log.info('Calculating responses for each sweep')        
        sweep_response = pd.DataFrame(index=self.stim_table.index.values, columns=np.array(range(self.numbercells+1)).astype(str))
        sweep_response.rename(columns={str(self.numbercells) : 'dx'}, inplace=True)
        for index, row in self.stim_table.iterrows():
            start = row['start'] - self.interlength
            end = row['start'] + self.sweeplength + self.interlength

            for nc in range(self.numbercells):
                temp = self.celltraces[nc,start:end]
                sweep_response[str(nc)][index] = 100*((temp/np.mean(temp[:self.interlength]))-1)
            sweep_response['dx'][index] = self.dxcm[start:end]   
        
        mean_sweep_response = sweep_response.applymap(do_mean)
        
        pval = sweep_response.applymap(do_p_value)
        return sweep_response, mean_sweep_response, pval            
        




import scipy.stats as st
import pandas as pd
import numpy as np
from allensdk.brain_observatory.stimulus_analysis import StimulusAnalysis

class LocallySparseNoise(StimulusAnalysis):    
    """ Perform tuning analysis specific to the locally sparse noise stimulus. 

    Parameters
    ----------
    data_set: BrainObservatoryNwbDataSet object
    
    nrows: int
       Number of rows in the stimulus template

    ncol: int
       Number of columns in the stimulus template
    """

    LSN_ON = 255
    LSN_OFF = 0
    LSN_GREY = 127
    LSN_OFF_SCREEN = 64

    def __init__(self, data_set, nrows=None, ncols=None, **kwargs):
        super(LocallySparseNoise, self).__init__(data_set, **kwargs)        
        self.nrows = 16 if nrows is None else nrows
        self.ncols = 28 if ncols is None else ncols

        self.stim_table = self.data_set.get_stimulus_table('locally_sparse_noise')
        self.LSN, self.LSN_mask = self.data_set.get_locally_sparse_noise_stimulus_template(mask_off_screen=False)
        self.sweeplength = self.stim_table['end'][1] - self.stim_table['start'][1]
        self.interlength = 4*self.sweeplength
        self.extralength = self.sweeplength
        self.sweep_response, self.mean_sweep_response, self.pval = self.get_sweep_response()
        self.receptive_field = self.get_receptive_field()

        
    def get_receptive_field(self):
        ''' Calculates receptive fields for each cell
        '''
        print "Calculating mean responses"
        receptive_field = np.empty((self.nrows, self.ncols, self.numbercells+1, 2))

        for xp in range(self.nrows):
            for yp in range(self.ncols):
                on_frame = np.where(self.LSN[:,xp,yp]==self.LSN_ON)[0]
                off_frame = np.where(self.LSN[:,xp,yp]==self.LSN_OFF)[0]
                subset_on = self.mean_sweep_response[self.stim_table.frame.isin(on_frame)]
                subset_off = self.mean_sweep_response[self.stim_table.frame.isin(off_frame)]
                receptive_field[xp,yp,:,0] = subset_on.mean(axis=0)
                receptive_field[xp,yp,:,1] = subset_off.mean(axis=0)
        return receptive_field  



# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.


import pytest
from mock import MagicMock
from allensdk.api.cache import Cache
from allensdk.api.queries.rma_api import RmaApi
import allensdk.core.json_utilities as ju
import pandas as pd
import pandas.io.json as pj


@pytest.fixture
def cache():
    return Cache()


@pytest.fixture
def rma():
    return RmaApi()


def test_wrap_json(rma, cache):
    msg = [ { 'whatever': True } ]
    
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={ 'msg': msg })
    ju.write = \
        MagicMock(name='write')
    
    ju.read = \
        MagicMock(name='read',
                  return_value=pd.DataFrame(msg))
           
    df = cache.wrap(rma.model_query,
                    'example.txt',
                    cache=True,
                    model='Hemisphere')
    
    assert df.loc[:,'whatever'][0]
    
    ju.read_url_get.assert_called_once_with(
        'http://api.brain-map.org/api/v2/data/query.json?q=model::Hemisphere')
    ju.write.assert_called_once_with('example.txt', msg)
    ju.read.assert_called_once_with('example.txt')


def test_wrap_dataframe(rma, cache):
    msg = [ { 'whatever': True } ]
     
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={ 'msg': msg })
    ju.write = \
        MagicMock(name='write')
    pj.read_json = \
        MagicMock(name='read_json',
                  return_value=msg)
     
    json_data = cache.wrap(rma.model_query,
                           'example.txt',
                           cache=True,
                           return_dataframe=True,
                           model='Hemisphere')
     
    assert json_data[0]['whatever']
    
    ju.read_url_get.assert_called_once_with(
        'http://api.brain-map.org/api/v2/data/query.json?q=model::Hemisphere')
    ju.write.assert_called_once_with('example.txt', msg)
    pj.read_json.assert_called_once_with('example.txt', orient='records')



# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import os
from allensdk.api.queries.rma_template import RmaTemplate
from allensdk.config.manifest import Manifest
import allensdk.brain_observatory.stimulus_info as stimulus_info
import logging

class BrainObservatoryApi(RmaTemplate):
    _log = logging.getLogger('allensdk.api.queries.brain_observatory_api')

    NWB_FILE_TYPE = 'NWBOphys'

    rma_templates = \
        {"brain_observatory_queries": [
            {'name': 'list_isi_experiments',
             'description': 'see name',
             'model': 'IsiExperiment',
             'num_rows': 'all',
             'count': False,
             'criteria_params': []
            },                                           
            {'name': 'isi_experiment_by_ids',
             'description': 'see name',
             'model': 'IsiExperiment',
             'criteria': '[id$in{{ isi_experiment_ids }}]',
             'include': 'experiment_container(ophys_experiments,targeted_structure)',
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['isi_experiment_ids']
            },
            {'name': 'ophys_experiment_by_ids',
             'description': 'see name',
             'model': 'OphysExperiment',
             'criteria': '{% if ophys_experiment_ids is defined %}[id$in{{ ophys_experiment_ids }}]{%endif%}',
             'include': 'well_known_files(well_known_file_type),targeted_structure,specimen(donor(age,transgenic_lines))',
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['ophys_experiment_ids']
             },                
            {'name': 'ophys_experiment_data',
             'description': 'see name',
             'model': 'WellKnownFile',
             'criteria': '[attachable_id$eq{{ ophys_experiment_id }}],well_known_file_type[name$eq%s]' % NWB_FILE_TYPE,
             'num_rows': 'all',
             'count': False,
             'criteria_params': [ 'ophys_experiment_id' ]
            },
            {'name': 'column_definitions',
             'description': 'see name',
             'model': 'ApiColumnDefinition',
             'criteria': '[api_class_name$eq{{ api_class_name }}]',
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['api_class_name']
            },
            {'name': 'column_definition_class_names',
             'description': 'see name',
             'model': 'ApiColumnDefinition',
             'only': ['api_class_name'],
             'num_rows': 'all',
             'count': False,
            },                                           
            {'name': 'stimulus_mapping',
             'description': 'see name',
             'model': 'ApiCamStimulusMapping',
             'criteria': '{% if stimulus_mapping_ids is defined %}[id$in{{ stimulus_mapping_ids }}]{%endif%}',
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['stimulus_mapping_ids']
            },
            {'name': 'experiment_container',
             'description': 'see name',
             'model': 'ExperimentContainer',
             'criteria': '{% if experiment_container_ids is defined %}[id$in{{ experiment_container_ids }}]{%endif%}',
             'include': 'ophys_experiments,isi_experiment,specimen(donor(age,transgenic_lines)),targeted_structure',
             'num_rows': 'all',
             'count': False, 
             'criteria_params': ['experiment_container_ids']
            },
            {'name': 'experiment_container_metric',
             'description': 'see name',
             'model': 'ApiCamExperimentContainerMetric',
             'criteria': '{% if experiment_container_metric_ids is defined %}[id$in{{ experiment_container_metric_ids }}]{%endif%}',
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['experiment_container_metric_ids']
            },
            {'name': 'cell_metric',
             'description': 'see name',
             'model': 'ApiCamCellMetric',
             'criteria': '{% if cell_specimen_ids is defined %}[cell_specimen_id$in{{ cell_specimen_ids }}]{%endif%}',
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['cell_specimen_ids']
            }
        ]}
    
    
    def __init__(self, base_uri=None):
        super(BrainObservatoryApi, self).__init__(base_uri,
                                                  query_manifest=BrainObservatoryApi.rma_templates)
    
    
    def get_ophys_experiments(self, ophys_experiment_ids=None):
        ''' Get OPhys Experiments by id
        
        Parameters
        ----------
        ophys_experiment_ids : integer or list of integers, optional
            only select specific experiments.
        
        Returns
        -------
        dict : ophys experiment metadata        
        '''
        data = self.template_query('brain_observatory_queries',
                                   'ophys_experiment_by_ids',
                                   ophys_experiment_ids=ophys_experiment_ids)
        
        return data
    
    
    def get_isi_experiments(self, isi_experiment_ids=None):
        ''' Get ISI Experiments by id
        
        Parameters
        ----------
        isi_experiment_ids : integer or list of integers, optional
            only select specific experiments.
        
        Returns
        -------
        dict : isi experiment metadata        
        '''
        data = self.template_query('brain_observatory_queries',
                                   'isi_experiment_by_ids',
                                   isi_experiment_ids=isi_experiment_ids)
        
        return data
    
    
    def list_isi_experiments(self, isi_ids=None):
        '''List ISI experiments available through the Allen Institute API
        
        Parameters
        ----------
        neuronal_model_ids : integer or list of integers, optional
            only select specific isi experiments.
        
        Returns
        -------
        dict : neuronal model metadata        
        '''
        data = self.template_query('brain_observatory_queries',
                                   'list_isi_experiments')
        
        return data


    def list_column_definition_class_names(self):
        ''' Get column definitions
        
        Parameters
        ----------
        
        Returns
        -------
        list : api class name strings        
        '''
        data = self.template_query('brain_observatory_queries',
                                   'column_definition_class_names')
        
        names = list(set([n['api_class_name'] for n in data]))
        
        return names
    
    
    def get_column_definitions(self, api_class_name=None):
        ''' Get column definitions
        
        Parameters
        ----------
        api_class_names : string or list of strings, optional
            only select specific column definition records.
        
        Returns
        -------
        dict : column definition metadata        
        '''
        data = self.template_query('brain_observatory_queries',
                                   'column_definitions',
                                   api_class_name=api_class_name)
        
        return data
    
    
    # TODO: search by item type and level
    def get_stimulus_mappings(self, stimulus_mapping_ids=None):
        ''' Get stimulus mappings by id
        
        Parameters
        ----------
        stimulus_mapping_ids : integer or list of integers, optional
            only select specific stimulus mapping records.
        
        Returns
        -------
        dict : stimulus mapping metadata        
        '''
        data = self.template_query('brain_observatory_queries',
                                   'stimulus_mapping',
                                   stimulus_mapping_ids=stimulus_mapping_ids)
        
        return data
    
    
    def get_cell_metrics(self, cell_specimen_ids=None):
        ''' Get cell metrics by id
        
        Parameters
        ----------
        cell_metrics_ids : integer or list of integers, optional
            only select specific cell metric records.
        
        Returns
        -------
        dict : cell metric metadata        
        '''
        self._log.warning("Downloading metrics and metadata for all cells. This can take some time.")

        data = self.template_query('brain_observatory_queries',
                                   'cell_metric',
                                   cell_specimen_ids=cell_specimen_ids)

        return data
    
    
    def get_experiment_containers(self, experiment_container_ids=None):
        ''' Get experiment container by id
        
        Parameters
        ----------
        experiment_container_ids : integer or list of integers, optional
            only select specific experiment containers.
        
        Returns
        -------
        dict : experiment container metadata        
        '''
        data = self.template_query('brain_observatory_queries',
                                   'experiment_container',
                                   experiment_container_ids=experiment_container_ids)
        
        return data
    
    
    def get_experiment_container_metrics(self, experiment_container_metric_ids=None):
        ''' Get experiment container metrics by id
        
        Parameters
        ----------
        isi_experiment_ids : integer or list of integers, optional
            only select specific experiments.
        
        Returns
        -------
        dict : isi experiment metadata        
        '''
        data = self.template_query('brain_observatory_queries',
                                   'experiment_container_metric',
                                   experiment_container_metric_ids=experiment_container_metric_ids)
        
        return data
    
    
    def save_ophys_experiment_data(self, ophys_experiment_id, file_name):
        dirname = os.path.dirname(file_name)
        Manifest.safe_mkdir(dirname)


        data = self.template_query('brain_observatory_queries',
                                   'ophys_experiment_data',
                                   ophys_experiment_id=ophys_experiment_id)
        
        try:
            file_url = data[0]['download_link']
        except Exception as _:
            raise Exception("ophys experiment %d has no data file" % ophys_experiment_id)

        self._log.warning("Downloading ophys_experiment %d NWB. This can take some time." % ophys_experiment_id)

        self.retrieve_file_over_http(self.api_url + file_url, file_name)
        
    
    def filter_experiment_containers(self, containers, 
                                     ids=None, 
                                     targeted_structures=None, 
                                     imaging_depths=None, 
                                     transgenic_lines=None):

        if ids is not None:
            containers = [ c for c in containers if c['id'] in ids ]

        if targeted_structures is not None:
            containers = [ c for c in containers if c['targeted_structure']['acronym'] in targeted_structures ]

        if imaging_depths is not None:
            containers = [ c for c in containers if c['imaging_depth'] in imaging_depths ]
        
        if transgenic_lines is not None:
            containers = [ c for c in containers for tl in c['specimen']['donor']['transgenic_lines'] if tl['name'] in transgenic_lines ]

        return containers
    
    
    def filter_ophys_experiments(self, experiments, 
                                 ids=None,
                                 experiment_container_ids=None,
                                 targeted_structures=None, 
                                 imaging_depths=None, 
                                 transgenic_lines=None, 
                                 stimuli=None, 
                                 session_types=None):

        # re-using the code from above
        experiments = self.filter_experiment_containers(experiments, 
                                                        ids=ids,
                                                        targeted_structures=targeted_structures, 
                                                        imaging_depths=imaging_depths, 
                                                        transgenic_lines=transgenic_lines)

        if experiment_container_ids is not None:
            experiments = [ e for e in experiments if e['experiment_container_id'] in experiment_container_ids ]
            
        if session_types is not None:
            experiments = [ e for e in experiments if e['stimulus_name'] in session_types ]

        if stimuli is not None:
            experiments = [ e for e in experiments 
                            if len(set(stimuli) & set(stimulus_info.stimuli_in_session(e['stimulus_name']))) > 0 ]

        return experiments

    def filter_cell_specimens(self, cell_specimens, ids=None, experiment_container_ids=None):
        if ids is not None:
            cell_specimens = [ c for c in cell_specimens if c['cell_specimen_id'] in ids ]

        if experiment_container_ids is not None:
            cell_specimens = [ c for c in cell_specimens if e['experiment_container_id'] in experiment_container_ids ]

        return cell_specimens



# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from stimulus_analysis import StimulusAnalysis
import scipy.stats as st
import pandas as pd
import numpy as np
from math import sqrt
import logging

class DriftingGratings(StimulusAnalysis):
    """ Perform tuning analysis specific to drifting gratings stimulus. 

    Parameters
    ----------
    data_set: BrainObservatoryNwbDataSet object
    """

    _log = logging.getLogger('allensdk.brain_observatory.drifting_gratings')

    def __init__(self, data_set, **kwargs):
        super(DriftingGratings, self).__init__(data_set, **kwargs)                   
        stimulus_table = self.data_set.get_stimulus_table('drifting_gratings')
        self.stim_table = stimulus_table.fillna(value=0.)     
        self.sweeplength = 60#self.sync_table['end'][1] - self.sync_table['start'][1]
        self.interlength = 30#self.sync_table['start'][2] - self.sync_table['end'][1]
        self.extralength = 0
        self.orivals = np.unique(self.stim_table.orientation).astype(int)
        self.tfvals = np.unique(self.stim_table.temporal_frequency).astype(int)
        self.number_ori = len(self.orivals)
        self.number_tf = len(self.tfvals)            
        self.sweep_response, self.mean_sweep_response, self.pval = self.get_sweep_response()
        self.response = self.get_response()
        self.peak = self.get_peak()
    
    def get_response(self):
        ''' Computes the mean response for each cell to each stimulus condition.  Return is 
        a (# orientations, # temporal frequencies, # cells, 3) np.ndarray.  The final dimension
        contains the mean response to the condition (index 0), standard error of the mean of the response
        to the condition (index 1), and p value of the response to that condition (index 3).

        Returns
        -------
        Numpy array storing mean responses.
        '''
        DriftingGratings._log.info("Calculating mean responses")
        
        response = np.empty((self.number_ori, self.number_tf, self.numbercells+1, 3))
        def ptest(x):
            return len(np.where(x<(0.05/(8*5)))[0])
            
        for ori in self.orivals:
            ori_pt = np.where(self.orivals == ori)[0][0]
            for tf in self.tfvals:
                tf_pt = np.where(self.tfvals == tf)[0][0]
                subset_response = self.mean_sweep_response[(self.stim_table.temporal_frequency==tf)&(self.stim_table.orientation==ori)]
                subset_pval = self.pval[(self.stim_table.temporal_frequency==tf)&(self.stim_table.orientation==ori)]
                response[ori_pt, tf_pt, :, 0] = subset_response.mean(axis=0)
                response[ori_pt, tf_pt, :, 1] = subset_response.std(axis=0)/sqrt(len(subset_response))
                response[ori_pt, tf_pt, :, 2] = subset_pval.apply(ptest, axis=0)
        return response
    
    def get_peak(self):
        ''' Computes metrics related to each cell's peak response condition.
        
        Returns
        -------
        Pandas data frame containing the following columns (_dg suffix is
        for drifting grating):
            * ori_dg (orientation)
            * tf_dg (temporal frequency)
            * response_reliability_dg
            * osi_dg (orientation selectivity index)
            * dsi_dg (direction selectivity index)
            * peak_dff_dg (peak dF/F)
            * ptest_dg
            * p_run_dg
            * run_modulation_dg
            * cv_dg (circular variance)
        '''
        DriftingGratings._log.info('Calculating peak response properties')
        
        peak = pd.DataFrame(index=range(self.numbercells), columns=('ori_dg','tf_dg','response_reliability_dg','osi_dg','dsi_dg','peak_dff_dg','ptest_dg', 'p_run_dg','run_modulation_dg','cv_dg','cell_specimen_id'))
        cids = self.data_set.get_cell_specimen_ids()

        orivals_rad = np.deg2rad(self.orivals)
        for nc in range(self.numbercells):
            cell_peak = np.where(self.response[:,1:,nc,0] == np.nanmax(self.response[:,1:,nc,0]))
            prefori = cell_peak[0][0]
            preftf = cell_peak[1][0]+1
            peak.cell_specimen_id.iloc[nc] = cids[nc]
            peak.ori_dg.iloc[nc] = prefori
            peak.tf_dg.iloc[nc] = preftf
            peak.response_reliability_dg.iloc[nc] = self.response[prefori, preftf, nc, 2]/0.15
            pref = self.response[prefori, preftf, nc, 0]            
            orth1 = self.response[np.mod(prefori+2, 8), preftf, nc, 0]
            orth2 = self.response[np.mod(prefori-2, 8), preftf, nc, 0]
            orth = (orth1+orth2)/2
            null = self.response[np.mod(prefori+4, 8), preftf, nc, 0]
            
            tuning = self.response[:, preftf, nc, 0]                
            CV_top = np.empty((8))
            for i in range(8):
                CV_top[i] = (tuning[i]*np.exp(1j*2*orivals_rad[i])).real
            peak.cv_dg.iloc[nc] = np.abs(CV_top.sum()/tuning.sum())
            
            peak.osi_dg.iloc[nc] = (pref-orth)/(pref+orth) 
            peak.dsi_dg.iloc[nc] = (pref-null)/(pref+null)
            peak.peak_dff_dg.iloc[nc] = pref
            
            groups = []
            for ori in self.orivals:
                for tf in self.tfvals[1:]:
                    groups.append(self.mean_sweep_response[(self.stim_table.temporal_frequency==tf)&(self.stim_table.orientation==ori)][str(nc)])
            groups.append(self.mean_sweep_response[self.stim_table.temporal_frequency==0][str(nc)])
            f,p = st.f_oneway(*groups)
            peak.ptest_dg.iloc[nc] = p
            
            subset = self.mean_sweep_response[(self.stim_table.temporal_frequency==self.tfvals[preftf])&(self.stim_table.orientation==self.orivals[prefori])]
            subset_stat = subset[subset.dx<1]
            subset_run = subset[subset.dx>=1]
            if (len(subset_run)>2) & ( len(subset_stat)>2):
                (f, peak.p_run_dg.iloc[nc]) = st.ks_2samp(subset_run[str(nc)], subset_stat[str(nc)])
                peak.run_modulation_dg.iloc[nc] = subset_run[str(nc)].mean()/subset_stat[str(nc)].mean()
            else:
                peak.p_run_dg.iloc[nc] = np.NaN
                peak.run_modulation_dg.iloc[nc] = np.NaN

        
        return peak
    



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 14 14:04:24 2015

@author: saskiad@alleninstitute.org
Functions to extract relevant data from the CAM NWB files
"""
import h5py
import pandas as pd
import numpy as np
import allensdk.brain_observatory.roi_masks as roi
import itertools
from collections import defaultdict
from allensdk.brain_observatory.locally_sparse_noise import LocallySparseNoise
import dateutil
import re

class MissingStimulusException(Exception): pass

class BrainObservatoryNwbDataSet(object):
    MOVIE_FOV_PX = (512, 512)
    PIPELINE_DATASET = 'brain_observatory_pipeline'
    
    FILE_METADATA_MAPPING = { 
        'age': 'general/subject/age',
        'sex': 'general/subject/sex',
        'imaging_depth': 'general/optophysiology/imaging_plane_1/imaging depth',
        'targeted_structure': 'general/optophysiology/imaging_plane_1/location',
        'ophys_experiment_id': 'general/session_id',
        'experiment_container_id': 'general/experiment_container_id',
        'device_string': 'general/devices/2-photon microscope',
        'excitation_lambda': 'general/optophysiology/imaging_plane_1/excitation_lambda',
        'indicator': 'general/optophysiology/imaging_plane_1/indicator',
        'fov': 'general/fov',
        'genotype': 'general/subject/genotype',
        'session_start_time': 'session_start_time',
        'session_type': 'general/session_type'
        }

    STIMULUS_TABLE_TYPES = {
        'abstract_feature_series': [ 'drifting_gratings', 'static_gratings' ],
        'indexed_time_series': [ 'natural_movie_one', 'natural_movie_two', 'natural_movie_three', 
                                 'natural_scenes', 'locally_sparse_noise' ]
        }
    
    def __init__(self, nwb_file):
        self.nwb_file = nwb_file
        

    def get_fluorescence_traces(self, cell_specimen_ids=None):
        ''' Returns an array of fluorescence traces for all ROI and 
        the timestamps for each datapoint

        Parameters
        ----------
        cell_specimen_ids: list or array (optional)
            List of cell IDs to return traces for. If this is None (default)
            then all are returned

        Returns
        -------
        timestamps: 2D numpy array
            Timestamp for each fluorescence sample

        traces: 2D numpy array
            Fluorescence traces for each cell
        '''
        all_cell_specimen_ids = list(self.get_cell_specimen_ids())

        with h5py.File(self.nwb_file, 'r') as f:
            timestamps = f['processing'][self.PIPELINE_DATASET]['Fluorescence']['imaging_plane_1']['timestamps'].value
            ds = f['processing'][self.PIPELINE_DATASET]['Fluorescence']['imaging_plane_1']['data']

            if cell_specimen_ids is None:
                cell_traces = ds.value 
            else:
                inds = [ all_cell_specimen_ids.index(i) for i in cell_specimen_ids ]
                cell_traces = ds[inds,:]

        return timestamps, cell_traces

    def get_neuropil_traces(self, cell_specimen_ids=None):
        ''' Returns an array of fluorescence traces for all ROIs 
        and the timestamps for each datapoint

        Parameters
        ----------
        cell_specimen_ids: list or array (optional)
            List of cell IDs to return traces for. If this is None (default)
            then all are returned

        Returns
        -------
        timestamps: 2D numpy array
            Timestamp for each fluorescence sample

        traces: 2D numpy array
            Neuropil fluorescence traces for each cell
        '''
        all_cell_specimen_ids = self.get_cell_specimen_ids()

        with h5py.File(self.nwb_file, 'r') as f:
            timestamps = f['processing'][self.PIPELINE_DATASET]['Fluorescence']['imaging_plane_1']['timestamps'].value

            ds = f['processing'][self.PIPELINE_DATASET]['Fluorescence']['imaging_plane_1']['neuropil_traces']
            if cell_specimen_ids is None:
                np_traces = ds.value 
            else:
                inds = [ list(all_cell_specimen_ids).index(i) for i in cell_specimen_ids ]
                np_traces = ds[inds,:]

        return timestamps, np_traces

    def get_corrected_fluorescence_traces(self, cell_specimen_ids=None):
        ''' Returns an array of neuropil-corrected fluorescence traces 
        for all ROIs and the timestamps for each datapoint

        Parameters
        ----------
        cell_specimen_ids: list or array (optional)
            List of cell IDs to return traces for. If this is None (default)
            then all are returned

        Returns
        -------
        timestamps: 2D numpy array
            Timestamp for each fluorescence sample

        traces: 2D numpy array
            Corrected fluorescence traces for each cell
        '''
        all_cell_specimen_ids = list(self.get_cell_specimen_ids())

        with h5py.File(self.nwb_file, 'r') as f:
            timestamps = f['processing'][self.PIPELINE_DATASET]['Fluorescence']['imaging_plane_1']['timestamps'].value
            cell_traces_ds = f['processing'][self.PIPELINE_DATASET]['Fluorescence']['imaging_plane_1']['data']
            np_traces_ds = f['processing'][self.PIPELINE_DATASET]['Fluorescence']['imaging_plane_1']['neuropil_traces']
            r_ds = f['processing'][self.PIPELINE_DATASET]['Fluorescence']['imaging_plane_1']['r']

            if cell_specimen_ids is None:
                cell_traces = cell_traces_ds.value 
                np_traces = np_traces_ds.value 
                r = r_ds.value 
            else:
                inds = [ all_cell_specimen_ids.index(i) for i in cell_specimen_ids ]
                cell_traces = cell_traces_ds[inds,:]
                np_traces = np_traces_ds[inds,:]
                r = r_ds[inds]

        fc = cell_traces - np_traces * r[:, np.newaxis]

        return timestamps, fc

        
    def get_dff_traces(self, cell_specimen_ids=None):
        ''' Returns an array of dF/F traces for all ROIs and 
        the timestamps for each datapoint

        Parameters
        ----------
        cell_specimen_ids: list or array (optional)
            List of cell IDs to return data for. If this is None (default)
            then all are returned

        Returns
        -------
        timestamps: 2D numpy array
            Timestamp for each fluorescence sample

        dF/F: 2D numpy array
            dF/F values for each cell
        '''
        all_cell_specimen_ids = list(self.get_cell_specimen_ids())

        with h5py.File(self.nwb_file, 'r') as f:
            timestamps = f['processing'][self.PIPELINE_DATASET]['DfOverF']['imaging_plane_1']['timestamps'].value

            if cell_specimen_ids is None:
                cell_traces = f['processing'][self.PIPELINE_DATASET]['DfOverF']['imaging_plane_1']['data'].value 
            else:
                inds = [ all_cell_specimen_ids.index(i) for i in cell_specimen_ids ]
                cell_traces = f['processing'][self.PIPELINE_DATASET]['DfOverF']['imaging_plane_1']['data'][inds,:]

        return timestamps, cell_traces


    def get_roi_ids(self):
        ''' Returns an array of IDs for all ROIs in the file

        Returns
        -------
        ROI IDs: list
        '''
        with h5py.File(self.nwb_file, 'r') as f:
            roi_id = f['processing'][self.PIPELINE_DATASET]['ImageSegmentation']['roi_ids'].value 
        return roi_id

    def get_cell_specimen_ids(self):
        ''' Returns an array of cell IDs for all cells in the file

        Returns
        -------
        cell specimen IDs: list
        '''
        with h5py.File(self.nwb_file, 'r') as f:
            cell_id = f['processing'][self.PIPELINE_DATASET]['ImageSegmentation']['cell_specimen_ids'].value 
        return cell_id

    def get_session_type(self):
        ''' Returns the type of experimental session, presently one of the
        following: three_session_A, three_session_B, three_session_C

        Returns
        -------
        session type: string
        '''
        with h5py.File(self.nwb_file, 'r') as f:
            session_type = f['general/session_type'].value
        return session_type

        
    def get_max_projection(self):
        '''Returns the maximum projection image for the 2P movie.
        
        Returns
        -------
        max projection: np.ndarray
        '''

        with h5py.File(self.nwb_file, 'r') as f:
            max_projection = f['processing'][self.PIPELINE_DATASET]['ImageSegmentation']['imaging_plane_1']['reference_images']['maximum_intensity_projection_image']['data'].value
        return max_projection

    def list_stimuli(self):
        ''' Return a list of the stimuli presented in the experiment. 
        
        Returns
        -------
        stimuli: list of strings
        '''

        with h5py.File(self.nwb_file, 'r') as f:
            keys = f["stimulus/presentation/"].keys()
        return [ k.replace('_stimulus','') for k in keys ]

    def get_stimulus_table(self, stimulus_name):
        ''' Return a stimulus table given a stimulus name '''
        if stimulus_name in self.STIMULUS_TABLE_TYPES['abstract_feature_series']:
            return self.get_abstract_feature_series_stimulus_table(stimulus_name + "_stimulus")
        elif stimulus_name in self.STIMULUS_TABLE_TYPES['indexed_time_series']:
            return self.get_indexed_time_series_stimulus_table(stimulus_name + "_stimulus")
        elif stimulus_name == 'spontaneous':
            return self.get_spontaneous_activity_stimulus_table()
        else:
            raise IOError("Could not find a stimulus table named '%s'" % stimulus_name)

    def get_spontaneous_activity_stimulus_table(self):
        ''' Return the spontaneous activity stimulus table, if it exists.

        Returns
        -------
        stimulus table: pd.DataFrame
        '''
        k = "stimulus/presentation/spontaneous_stimulus"
        with h5py.File(self.nwb_file, 'r') as f:
            events = f[k + '/data'].value
            frame_dur = f[k + '/frame_duration'].value

        start_inds = np.where(events == 1)
        stop_inds = np.where(events == -1)

        if len(start_inds) != len(stop_inds):
            raise Exception("inconsistent start and time times in spontaneous activity stimulus table")

        stim_data = np.column_stack([frame_dur[start_inds,0].T, frame_dur[stop_inds,0].T]).astype(int)

        stimulus_table = pd.DataFrame(stim_data, columns=['start','end'])

        return stimulus_table

    def get_indexed_time_series_stimulus_table(self, stimulus_name):
        ''' Return the a stimulus table for an indexed time series.

        Returns
        -------
        stimulus table: pd.DataFrame
        '''
        
        k = "stimulus/presentation/%s" % stimulus_name

        with h5py.File(self.nwb_file, 'r') as f: 
            if k not in f:
                raise MissingStimulusException("Stimulus not found: %s" % stimulus_name)    
            inds = f[k + '/data'].value
            frame_dur = f[k + '/frame_duration'].value

        stimulus_table = pd.DataFrame(inds, columns=['frame'])
        stimulus_table.loc[:,'start'] = frame_dur[:,0].astype(int)
        stimulus_table.loc[:,'end'] = frame_dur[:,1].astype(int)

        return stimulus_table

    def get_abstract_feature_series_stimulus_table(self, stimulus_name):
        ''' Return the a stimulus table for an abstract feature series.

        Returns
        -------
        stimulus table: pd.DataFrame
        '''

        k = "stimulus/presentation/%s" % stimulus_name

        with h5py.File(self.nwb_file, 'r') as f:
            if k not in f:
                raise MissingStimulusException("Stimulus not found: %s" % stimulus_name)
            stim_data = f[k + '/data'].value
            features = f[k + '/features'].value
            frame_dur = f[k + '/frame_duration'].value

        stimulus_table = pd.DataFrame(stim_data, columns=features)
        stimulus_table.loc[:,'start'] = frame_dur[:,0].astype(int)
        stimulus_table.loc[:,'end'] = frame_dur[:,1].astype(int)

        return stimulus_table

    def get_stimulus_template(self, stimulus_name):
        ''' Return an array of the stimulus template for the specified stimulus.

        Parameters
        ----------
        stimulus_name: string
            Must be one of the strings returned by list_stimuli().

        Returns
        -------
        stimulus table: pd.DataFrame
        '''
        stim_name = stimulus_name + "_image_stack"
        with h5py.File(self.nwb_file, 'r') as f:
            image_stack = f['stimulus']['templates'][stim_name]['data'].value
        return image_stack

    def get_locally_sparse_noise_stimulus_template(self, mask_off_screen=True):
        ''' Return an array of the stimulus template for the specified stimulus.

        Parameters
        ----------
        mask_off_screen: boolean
           Set off-screen regions of the stimulus to LocallySparseNoise.LSN_OFF_SCREEN.

        Returns
        -------
        tuple: (template, off-screen mask)
        '''
        
        template = self.get_stimulus_template("locally_sparse_noise")
        
        # build mapping from template coordinates to display coordinates
        template_shape = (28, 16)
        template_display_shape = (1260, 720)
        display_shape = (1920, 1200)

        scale = [
            float(template_shape[0]) / float(template_display_shape[0]),
            float(template_shape[1]) / float(template_display_shape[1])
            ]
        offset = [
            -(display_shape[0] - template_display_shape[0]) * 0.5,
             -(display_shape[1] - template_display_shape[1]) * 0.5
             ]

        x,y = np.meshgrid(np.arange(display_shape[0]), np.arange(display_shape[1]), indexing='ij')
        template_display_coords = np.array([(x + offset[0]) * scale[0] - 0.5, 
                                            (y + offset[1]) * scale[1] - 0.5], 
                                           dtype=float)
        template_display_coords = np.rint(template_display_coords).astype(int)

        # build mask
        template_mask, template_frac = mask_stimulus_template(template_display_coords, template_shape)

        if mask_off_screen:
            template[:,~template_mask.T] = LocallySparseNoise.LSN_OFF_SCREEN

        return template, template_mask.T
    
    def get_roi_mask(self, cell_specimen_ids=None):
        ''' Returns an array of all the ROI masks

        Parameters
        ----------
        cell specimen IDs: list or array (optional)
            List of cell IDs to return traces for. If this is None (default)
            then all are returned

        Returns
        -------
            List of ROI_Mask objects
        '''

        all_cell_specimen_ids = self.get_cell_specimen_ids()

        with h5py.File(self.nwb_file, 'r') as f:
            mask_loc = f['processing'][self.PIPELINE_DATASET]['ImageSegmentation']['imaging_plane_1']
            roi_list = f['processing'][self.PIPELINE_DATASET]['ImageSegmentation']['imaging_plane_1']['roi_list'].value
        
            inds = None
            if cell_specimen_ids is None:
                inds = range(len(all_cell_specimen_ids))
            else:
                inds = [ list(all_cell_specimen_ids).index(i) for i in cell_specimen_ids ]

            roi_array = []
            for i in inds:
                v = roi_list[i]
                m = roi.create_roi_mask(self.MOVIE_FOV_PX[0], self.MOVIE_FOV_PX[1], [0,0,0,0], pix_list=mask_loc[v]["pix_mask"].value, label=v)
                roi_array.append(m)

        return roi_array
    
    def get_metadata(self):
        ''' Returns a dictionary of meta data associated with each 
        experiment, including Cre line, specimen number, 
        visual area imaged, imaging depth

        Returns
        -------
        metadata: dictionary
        '''
        
        meta = {}
            
        with h5py.File(self.nwb_file, 'r') as f:
            for memory_key, disk_key in BrainObservatoryNwbDataSet.FILE_METADATA_MAPPING.items():
                v = f[disk_key].value
                if v.dtype.type is np.string_:
                    v = str(v)
                meta[memory_key] = v

        meta['cre_line'] = meta['genotype'].split(';')[0]
        meta['imaging_depth_um'] = int(meta['imaging_depth'].split()[0])
        del meta['imaging_depth']
        meta['ophys_experiment_id'] = int(meta['ophys_experiment_id'])
        meta['experiment_container_id'] = int(meta['experiment_container_id'])
        meta['session_start_time'] = dateutil.parser.parse(meta['session_start_time'])

        # parse the age in days
        m = re.match("(.*?) days", meta['age'])
        if m:
            meta['age_days'] = int(m.groups()[0])
            del meta['age']
        else:
            raise IOError("Could not find device.")

        # parse the device string (ugly, sorry)
        device_string = meta['device_string']
        del meta['device_string']

        m = re.match("(.*?)\.\s(.*?)\sPlease*", device_string)
        if m:
            device, device_name = m.groups()
            meta['device'] = device
            meta['device_name'] = device_name
        else:
            raise IOError("Could not find device.")

        return meta
        

    
    def get_running_speed(self):
        ''' Returns the mouse running speed in cm/s
        '''
        with h5py.File(self.nwb_file, 'r') as f:
            dxcm = f['processing'][self.PIPELINE_DATASET]['BehavioralTimeSeries']['running_speed']['data'].value
            dxtime = f['processing'][self.PIPELINE_DATASET]['BehavioralTimeSeries']['running_speed']['timestamps'].value
            timestamps = f['processing'][self.PIPELINE_DATASET]['Fluorescence']['imaging_plane_1']['timestamps'].value    

        dxcm = dxcm[:,0]
        if dxtime[0] != timestamps[0]:
            adjust = np.where(timestamps==dxtime[0])[0][0]
            dxtime = np.insert(dxtime, 0, timestamps[:adjust])
            dxcm = np.insert(dxcm, 0, np.repeat(np.NaN, adjust))
        adjust = len(timestamps) - len(dxtime)
        if adjust>0:
            dxtime = np.append(dxtime, timestamps[(-1*adjust):])
            dxcm = np.append(dxcm, np.repeat(np.NaN, adjust))
        return dxcm, dxtime
    
    def get_motion_correction(self):
        ''' Returns a Panda DataFrame containing the x- and y- translation of each image used for image alignment
        '''
        with h5py.File(self.nwb_file, 'r') as f:
            motion_log = f['processing'][self.PIPELINE_DATASET]['MotionCorrection']['2p_image_series']['xy_translations']['data'].value
            motion_time = f['processing'][self.PIPELINE_DATASET]['MotionCorrection']['2p_image_series']['xy_translations']['timestamps'].value
            motion_names = f['processing'][self.PIPELINE_DATASET]['MotionCorrection']['2p_image_series']['xy_translations']['feature_description'].value
            motion_correction = pd.DataFrame(motion_log, columns=motion_names)
            motion_correction['timestamp'] = motion_time    

        return motion_correction
    
    
    def save_analysis_dataframes(self, *tables):
        # NOTE: should use NWB library to write data to NWB file. It is 
        #   designed to avoid possible corruption of the file in event
        #   of a failed write
        store = pd.HDFStore(self.nwb_file, mode='a')

        for k,v in tables:
            store.put('analysis/%s' % (k), v)

        store.close()
        

    def save_analysis_arrays(self, *datasets):    
        # NOTE: should use NWB library to write data to NWB file. It is 
        #   designed to avoid possible corruption of the file in event
        #   of a failed write
        with h5py.File(self.nwb_file, 'a') as f:
            for k,v in datasets:
                if k in f['analysis']:
                    del f['analysis'][k]
                f.create_dataset('analysis/%s' % k, data=v)


def warp_stimulus_coords(vertices,
                         distance=15.0,
                         mon_height_cm=32.5,
                         mon_width_cm=51.0,
                         mon_res=(1920,1200),
                         eyepoint=(0.5,0.5)):
    """  
    For a list of screen vertices, provides a corresponding list of texture coordinates.

    Parameters
    ----------
    vertices: numpy.ndarray
        [[x0,y0], [x1,y1], ...] A set of vertices to  convert to texture positions.
    distance: float
        distance from the monitor in cm.
    mon_height_cm: float
        monitor height in cm
    mon_width_cm: float
        monitor width in cm
    mon_res: tuple
        monitor resolution (x,y)
    eyepoint: tuple

    Returns
    -------
    np.ndarray
        x,y coordinates shaped like the input that describe what pixel coordinates
        are displayed an the input coordinates after warping the stimulus.
    """

    mon_width_cm = float(mon_width_cm)
    mon_height_cm = float(mon_height_cm)
    distance = float(distance)
    mon_res_x, mon_res_y = float(mon_res[0]), float(mon_res[1])

    vertices = vertices.astype(np.float)

    # from pixels (-1920/2 -> 1920/2) to stimulus space (-0.5->0.5)
    vertices[:, 0] = vertices[:, 0] / mon_res_x
    vertices[:, 1] = vertices[:, 1] / mon_res_y

    x = (vertices[:,0] + 0.5) * mon_width_cm
    y = (vertices[:,1] + 0.5) * mon_height_cm

    xEye = eyepoint[0] * mon_width_cm
    yEye = eyepoint[1] * mon_height_cm

    x = x - xEye
    y = y - yEye

    r = np.sqrt(np.square(x) + np.square(y) + np.square(distance))

    azimuth = np.arctan(x / distance)
    altitude = np.arcsin(y / r)

    # calculate the texture coordinates
    tx = distance * (1 + x / r) - distance
    ty = distance * (1 + y / r) - distance

    # prevent div0
    azimuth[azimuth == 0] = np.finfo(np.float32).eps
    altitude[altitude == 0] = np.finfo(np.float32).eps

    # the texture coordinates (which are now lying on the sphere)
    # need to be remapped back onto the plane of the display.
    # This effectively stretches the coordinates away from the eyepoint.

    centralAngle = np.arccos(np.cos(altitude) * np.cos(np.abs(azimuth)))
    # distance from eyepoint to texture vertex
    arcLength = centralAngle * distance
    # remap the texture coordinate
    theta = np.arctan2(ty, tx)
    tx = arcLength * np.cos(theta)
    ty = arcLength * np.sin(theta)

    u_coords = tx / mon_width_cm 
    v_coords = ty / mon_height_cm

    retCoords = np.column_stack((u_coords, v_coords))

    # back to pixels
    retCoords[:, 0] = retCoords[:, 0] * mon_res_x
    retCoords[:, 1] = retCoords[:, 1] * mon_res_y

    return retCoords

def make_display_mask(display_shape=(1920,1200)):
    ''' Build a display-shaped mask that indicates which pixels are on screen after warping the stimulus. '''
    x = np.array(range(display_shape[0])) - display_shape[0]/2
    y = np.array(range(display_shape[1])) - display_shape[1]/2
    display_coords = np.array(list(itertools.product(x,y)))
                                                     
    warped_coords = warp_stimulus_coords(display_coords).astype(int)

    off_warped_coords = np.array([ warped_coords[:,0] + display_shape[0]/2,
                                   warped_coords[:,1] + display_shape[1]/2 ])

    
    used_coords = set()
    for i in range(off_warped_coords.shape[1]):
        used_coords.add((off_warped_coords[0,i], off_warped_coords[1,i]))

    used_coords = ( np.array([x for (x,y) in used_coords ]),
                    np.array([y for (x,y) in used_coords ]) )

    mask = np.zeros(display_shape)
    mask[used_coords] = 1

    return mask

def mask_stimulus_template(template_display_coords, template_shape, display_mask=None, threshold=1.0):
    ''' Build a mask for a stimulus template of a given shape and display coordinates that indicates
    which part of the template is on screen after warping.

    Parameters
    ----------
    template_display_coords: list
        list of (x,y) display coordinates

    template_shape: tuple
        (width,height) of the display template  

    display_mask: np.ndarray
        boolean 2D mask indicating which display coordinates are on screen after warping.

    threshold: float
        Fraction of pixels associated with a template display coordinate that should remain
        on screen to count as belonging to the mask. 

    Returns
    -------
    tuple: (template mask, pixel fraction)
    '''
    if display_mask is None:
        display_mask = make_display_mask()

    frac = np.zeros(template_shape)
    mask = np.zeros(template_shape, dtype=bool)
    for y in range(template_shape[1]):
        for x in range(template_shape[0]):
            tdcm = np.where((template_display_coords[0,:,:] == x) & (template_display_coords[1,:,:] == y))
            v = display_mask[tdcm]
            f = np.sum(v) / len(v)
            frac[x,y] = f
            mask[x,y] = f >= threshold
    
    return mask, frac

                      

    



""" The methods in this module are used for configuring dynamics and reset rules for the GlifNeuron.  
For more details on how to use these methods, see :doc:`glif_models`.
"""
import logging
import functools
import numpy as np
from numpy.distutils.npy_pkg_config import VariableSet

#fast_threshold=[]
#slow_threshold=[]

class GlifNeuronMethod( object ):
    """ A simple class to keep track of the name and parameters associated with a neuron method.
    This class is initialized with a name, function, and parameters to pass to the function.  The
    function then has those passed parameters fixed to a partial function using functools.partial.  
    This class then mimics a function itself using the __call__ convention. Parameters that are not 
    fixed in this way are assumed to be passed into the method when it is called.  If the passed
    parameters contain an argument that is not part of the function signature, an exception will
    be raised.    

    Parameters
    ----------
    method_name : string
        A shorthand name that will be used to reference this method in the `GlifNeuron`.
    method : function
        A python function to be called when this instance is called.
    method_params : dict
        A dictionary mapping function arguments to values for values that should be fixed.
    """

    def __init__(self, method_name, method, method_params):
        self.name = method_name
        self.params = method_params
        self.method = functools.partial(method, **method_params)

    def __call__(self, *args, **kwargs):
        """ Defining this method allows an instance to be called like a function """
        return self.method(*args, **kwargs)

    def to_dict(self):
        return {
            'name': self.name,
            'params': self.params
        }

    def modify_parameter(self, param, operator):
        """ Modify a function parameter needs to be modified after initialization. 

        Parameters
        ----------
        param : string
            the name of the parameter to modify
        operator : callable
            a function or lambda that returns the desired modified value

        Returns
        -------
        type
            the new value of the variable that was just modified.
        """
        value = operator(self.method.keywords[param])
        self.method.keywords[param] = value
        return value


def max_of_line_and_const(x,b,c,d):
    #TODO: move to other library
    """ Find the maximum of a value and a position on a line 

    Parameters
    ----------
    x: float
        x position on line 1
    c: float
        slope of line 1
    d: float
        y-intercept of line 1
    b: float
        y-intercept of line 2

    Returns
    -------
    float
        the max of a line value and a constant
    """

    one = b
    two = c*x+d
    return np.maximum(one,two)

#commenting this out to make sure it doest get called and I forgot to switch is somewhere
#def min_of_line_and_const(x,b,c,d):
#    #TODO: move to other library
#    """ Find the minimum of a value and a position on a line 
#
#    Parameters
#    ----------
#    x: float
#        x position on line 1
#    c: float
#        slope of line 1
#    d: float
#        y-intercept of line 1
#    b: float
#        y-intercept of line 2
#
#    Returns
#    -------
#    float
#        the max of a line value and a constant
#    """
#
#    one = b
#    two = c*x+d
#    return np.minimum(one,two)

def min_of_line_and_zero(x,c,d):
    #TODO: move to other library
    """ Find the minimum of a value and a position on a line 

    Parameters
    ----------
    x: float
        x position on line 1
    c: float
        slope of line 1
    d: float
        y-intercept of line 1
    b: float
        y-intercept of line 2

    Returns
    -------
    float
        the max of a line value and a constant
    """

    one = 0
    two = c*x+d
    return np.minimum(one,two)


def dynamics_AScurrent_exp(neuron, AScurrents_t0, time_step, spike_time_steps):
    """ Exponential afterspike current dynamics method takes a current at t0 and returns the current at
    a time step later.
    """

#     print 'HERE: dynamics_threshold_inf'
#    return AScurrents_t0*(1.0 - neuron.k*neuron.dt)
    return AScurrents_t0*np.exp(-neuron.k*neuron.dt)
        

def dynamics_AScurrent_vector(neuron, AScurrents_t0, time_step, spike_time_steps, vector):
    """ The vector afterspike current dynamics method keeps track of all of the afterspike
    currents for every previous spike and updates them based on the current time step.
    This method is very slow.

    Parameters
    ----------
    vector : np.ndarray
        an array of all running afterspike current values.
    """
    # a hack to convert lists into numpy arrays
    if isinstance(vector, list):
        vector = neuron.AScurrent_dynamics_method.modify_parameter('vector', np.array)
        
    total = np.zeros(len(vector))
        
    # run through all of the spikes, adding ascurrents based on how long its been since the spike occurred
    for spike_time_step in spike_time_steps:
        try:
            total += vector[:, time_step - spike_time_step]
        except Exception as e:
            pass
        
    return total
    

def dynamics_AScurrent_none(neuron, AScurrents_t0, time_step, spike_time_steps):
    """ This method always returns zeros for the afterspike currents, regardless of input. """
    return np.zeros(len(AScurrents_t0))


def dynamics_voltage_linear_forward_euler(neuron, voltage_t0, AScurrents_t0, inj):
    """ (TODO) Linear voltage dynamics. """
    return voltage_t0 + (inj + np.sum(AScurrents_t0) - neuron.G * neuron.coeffs['G'] * (voltage_t0 - neuron.El)) * neuron.dt / (neuron.C * neuron.coeffs['C'])

def dynamics_voltage_linear_exact(neuron, voltage_t0, AScurrents_t0, inj):
    """ (TODO) Linear voltage dynamics. """

    C = (neuron.C * neuron.coeffs['C'])
    I = inj + np.sum(AScurrents_t0)
    g = neuron.G * neuron.coeffs['G']
    tau = g/C
    N = (I+ g*neuron.El)/C

    return voltage_t0*np.exp(-neuron.dt*tau) + N*(1-np.exp(-tau*neuron.dt))/tau

def dynamics_voltage_piecewise_linear_exact(neuron, voltage_t0, AScurrents_t0, inj, R_tlparam1, R_tlparam2, R_t1param3, El_slope_param, El_intercept_param):
    """ (TODO) Linear voltage dynamics. """

    C = (neuron.C * neuron.coeffs['C'])
    I = inj + np.sum(AScurrents_t0)
    g = neuron.coeffs['G']/max_of_line_and_const(voltage_t0, R_tlparam1, R_tlparam2, R_t1param3)
    tau = g/C
    El = min_of_line_and_zero(voltage_t0, El_slope_param, El_intercept_param)
    N = (I+ g*El)/C
    
    G = 1. / (max_of_line_and_const(voltage_t0, R_tlparam1, R_tlparam2, R_t1param3))
    El = min_of_line_and_zero(voltage_t0, El_slope_param, El_intercept_param)

    return voltage_t0*np.exp(-neuron.dt*tau) + N*(1-np.exp(-tau*neuron.dt))/tau
    

def dynamics_voltage_quadratic_forward_euler(neuron, voltage_t0, AScurrents_t0, inj, a, b, c, d, e):    
    """ (TODO) Quadratic voltage dynamics equation.

    Parameters
    ----------
    a : float
        constant coefficient of voltage equation
    b : float
        linear coefficient of voltage equation
    c : float
        quadratic coefficient of voltage equation
    d : float
        voltage threshold
    e : float
        voltage used if voltage surpasses threshold (d)
    """
    I_of_v = a + b * voltage_t0 + c * voltage_t0**2  
        
    if voltage_t0 > d:
        I_of_v=e

    return voltage_t0 + (inj + np.sum(AScurrents_t0) - I_of_v) * neuron.dt / (neuron.C * neuron.coeffs['C']) 


def dynamics_voltage_piecewise_linear_forward_euler(neuron, voltage_t0, AScurrents_t0, inj, R_tlparam1, R_tlparam2, R_t1param3, El_slope_param, El_intercept_param):
    """ Piecewise linear voltage dynamics methods. This method requires 3 parameters for equations of both resistance and resting potential. 

    Parameters
    ----------
    R_tlparam1 : float 
        TODO
    R_tlparam2 : float
        TODO
    R_tlparam3 : float
        TODO
    El_tlparam1 : float
        TODO
    El_tlparam2 : float
        TODO
    El_tlparam3 : float
        TOOD
    """
    #confirm this is being implimented correctly.
    G = 1. / (max_of_line_and_const(voltage_t0, R_tlparam1, R_tlparam2, R_t1param3))
#    El = min_of_line_and_const(voltage_t0, El_tlparam1, El_tlparam2, El_t1param3) #this is for when El wasnt forced to zero
    El = min_of_line_and_zero(voltage_t0, El_slope_param, El_intercept_param)

    return voltage_t0 + (inj + np.sum(AScurrents_t0) - G * neuron.coeffs['G'] * (voltage_t0 - El)) * neuron.dt / (neuron.C * neuron.coeffs['C'])
    

def dynamics_threshold_adapt_standard(neuron, threshold_t0, voltage_t0, AScurrents_t0, inj, a, b):
    """ Standard adapting threshold dynamics equation.

    Parameters
    ----------
    neuron : class
    a : float
        coefficient of voltage
    b : float
        coefficient of threshold
    AScurrents_t0 : not used here 
    inj : not used here
    """
#     print 'HERE: dynamics_threshold_adapt_standard'
    return threshold_t0 + (a * neuron.coeffs['a'] * (voltage_t0-neuron.El) - 
                           b * neuron.coeffs['b'] * (threshold_t0 - neuron.coeffs['th_inf'] * neuron.th_inf)) * neuron.dt 

#TODO: depricate when sure that dynamics_threshold_three_sep_components will be the official function
#def dynamics_threshold_adapt_slow_plus_fast(neuron, threshold_t0, voltage_t0):        
#    """The threshold will adapt via two mechanisms: 1. a slower voltage dependent adaptation as in the dynamics_threshold_adapt_standard. 
#    These are the components which are fit via optimization and inditial conditions are supplied via the GLM. 2. A fast component initiated 
#    by a spike which quickly decays.  These values are estimated via the multi short square stimuli. 
#    """
#    
#    md = neuron.update_method_data
#
#    # initial conditions
#    if 'th_spike' not in md:
#        md['th_spike'] = [ 0 ]
#    if 'th_voltage' not in md:
#        md['th_voltage'] = [ neuron.th_inf * neuron.coeffs['th_inf'] ]
#
#    th_spike = md['th_spike'][-1]
#    th_voltage = md['th_voltage'][-1] 
#
##     print 'a_spike', md['a_spike']
##     print 'b_spike', md['b_spike']
##     print 'a_voltage', md['a_voltage']
##     print 'b_voltage', md['b_voltage']
#
#    voltage_component = th_voltage + ( md['a_voltage'] * neuron.coeffs['a'] * ( voltage_t0 - neuron.El ) - 
#                                       md['b_voltage'] * neuron.coeffs['b'] * ( th_voltage - neuron.coeffs['th_inf'] * neuron.th_inf ) ) * neuron.dt
#    spike_component = th_spike - md['b_spike'] * th_spike * neuron.dt
#
#    
#    #------hack to plot slow and fast component
##    global fast_threshold
##    global slow_threshold
##    fast_threshold.append(spike_component)
##    slow_threshold.append(voltage_component)
#    
#    #------update the voltage and spiking values of the the
#    md['th_spike'].append(spike_component)
#    md['th_voltage'].append(voltage_component)
#    
#    return voltage_component+spike_component

#TODO: deprecate when confirm the forward and exact equations are doing the right thing.
# def dynamics_threshold_three_sep_components_hybrid(neuron, threshold_t0, voltage_t0, AScurrents_t0, inj):        
#     """This is the sequel to the dynamics_threshold_adapt_slow_plus_fast module. Here as in 
#     dynamics_threshold_adapt_slow_plus_fast. The threshold will adapt via two mechanisms: 1. a 
#     slower voltage dependent adaptation as in the dynamics_threshold_adapt_standard. and 2. a 
#     fast component initiated by a spike which quickly decays.  These two component are 
#     in reference to zero and are recorded in the md parameters.  The third component refers to 
#     th_inf which is added on in the end as opposed to being encompassed in the voltage component 
#     of the threshold which caused a problem in optimizing th_inf.
#     AScurrents_t0 : not used here 
#     inj : not used here
#     """
#     md = neuron.update_method_data
# 
#     # initial conditions
#     if 'th_spike' not in md:
#         md['th_spike'] = [ 0 ]
#         logging.warning('delete this function it is a hybrid')
#     if 'th_voltage' not in md:
#         md['th_voltage'] = [ 0 ]
# 
#     th_spike = md['th_spike'][-1]
#     th_voltage = md['th_voltage'][-1] 
# 
#     #why are this variables being saved in an md variable?
#     #TODO: HOW CAN NEURON.COEFFS['A'] BE MULTIPLIED AFTERWARD SINCE IT IS ALSO MULTIPLYING THE 
#     #VOLTAGE COMPONENT.
#     voltage_component = th_voltage + ( md['a_voltage'] *  ( voltage_t0 - neuron.El ) - 
#                                        md['b_voltage'] * neuron.coeffs['b'] * ( th_voltage ) ) * neuron.dt
#     
#     #spike_component = th_spike - md['b_spike'] * th_spike * neuron.dt
#     spike_component = th_spike*np.exp(-md['b_spike'] * neuron.dt)
# 
#     #------update the voltage and spiking values of the the
#     md['th_spike'].append(spike_component)
#     md['th_voltage'].append(voltage_component)
#     
#     #TODO: I dont trust the a_coeff being multiplied by the whole function
#     return neuron.coeffs['a']*voltage_component+spike_component+neuron.th_inf * neuron.coeffs['th_inf']    

def spike_component_of_threshold_forward_euler(th_t0, b_spike, dt):
    '''Spike component of threshold modeled as an exponential decay. Implemented 
    here for forward Euler 
    
    Parameters
    ----------
    th_t0 : float
        threshold input to function
    b_spike : float
        decay constant of exponential
    dt : float
        time step
    '''
    b_spike=-b_spike #TODO: this is here because b_spike is always input as positive although it is negative
    return th_t0 + th_t0*b_spike * dt                                                                         
    
def spike_component_of_threshold_exact(th0, b_spike, t):
    '''Spike component of threshold modeled as an exponential decay. Implemented 
    here as exact analytical solution. 
    
    Parameters
    ----------
    th0 : float
        threshold input to function
    b_spike : float
        decay constant of exponential
    t : float or array
        time step if used in an Euler setup
        time if used analytically
    '''
    b_spike=-b_spike
    return th0*np.exp(b_spike * t)

def voltage_component_of_threshold_forward_euler(th_t0, v_t0, dt, a_voltage, b_voltage, El):
    '''Equation 2.1 of Mihalas and Nieber, 2009 implemented for use in forward Euler. Note 
    here all variables are in reference to threshold infinity.  Therefore thr_inf is zero 
    here (replaced threshold_inf with 0 in the equation to be verbose). This is done so that 
    th_inf can be optimized without affecting this function. 
    
    Parameters
    ----------
    th_t0 : float
        threshold input to function
    v_t0 : float
        voltage input to function
    dt : float
        time step
    a_voltage : float
        constant a
    b_voltage : float
        constant b
    El : float
        reversal potential
    '''
    return th_t0 + (a_voltage*(v_t0-El)-b_voltage*(th_t0-0))*dt

def voltage_component_of_threshold_exact(th0, v0, I, t, a_voltage, b_voltage, C, g, El):
    '''Note this function is the exact formulation; however, dt is used because t0 is the initial time and dt
    is the time the function is exactly evaluated at. Note: that here, this equation is in reference to th_inf.
    Therefore th0 is the total threshold-thr_inf (threshold_inf replaced with 0 in the equation to be verbose).  
    This is done so that th_inf can be optimized without affecting this function.
    
    Parameters
    ----------
    th0 : float
        threshold input to function
    v0 : float
        voltage input to function
    I : float
        total current entering neuron (note if there are after spike currents these must be included in this value)
    t : float or array
        time step if used in an Euler setup
        time if used analytically
    a_voltage : float
        constant a
    b_voltage : float
        constant b
    C : float
        capacitance
    g : float
        conductance (1/resistance)
    El : float 
        reversal potential
    '''
    beta=(I+g*El)/g
    phi=a_voltage/(b_voltage-g/C)
    return phi*(v0-beta)*np.exp(-g*t/C)+1/(np.exp(b_voltage*t))*(th0-phi*(v0-beta)-
                        (a_voltage/b_voltage)*(beta-El)-0) +(a_voltage/b_voltage)*(beta-El) +0


def dynamics_threshold_three_components_forward(neuron, threshold_t0, voltage_t0, AScurrents_t0, inj, 
                                                a_spike, b_spike, a_voltage, b_voltage):
    """Threshold dynamics implemented for forward Euler. The threshold will adapt via two mechanisms: 
    1. a voltage dependent adaptation.  
    2. a component initiated by a spike which decays as an exponential.  
    These two component are in reference to threshold infinity and are recorded 
    in the neuron's threshold components.
    The third component refers to th_inf which is added separately as opposed to being 
    included in the voltage component of the threshold as is done in equation 2.1 of
    Mihalas and Nieber 2009.  Threshold infinity is removed for simple optimization.
    
    Parameters
    ----------
    neuron : class
    threshold_t0 : float
        threshold input to function
    voltage_t0 : float
        voltage input to function
    AScurrents_t0 : not used here 
    inj : not used here
    """
    tcs = neuron.get_threshold_components()
    
    th_spike = tcs['spike'][-1]
    th_voltage = tcs['voltage'][-1] 

    a_voltage = a_voltage * neuron.coeffs['a']
    b_voltage = b_voltage * neuron.coeffs['b']

    spike_component = spike_component_of_threshold_forward_euler(th_t0, b_spike, neuron.dt)    
    voltage_component = voltage_component_of_threshold_forward_euler(th_voltage, voltage_t0, dt, a_voltage, b_voltage, neuron.El)

    neuron.add_threshold_components(spike_component, voltage_component)
    
    return voltage_component+spike_component+neuron.th_inf * neuron.coeffs['th_inf']    


def dynamics_threshold_three_components_exact(neuron, threshold_t0, voltage_t0, AScurrents_t0, inj,
                                              a_spike, b_spike, a_voltage, b_voltage):
    """Analitical solution for threshold dynamics. The threshold will adapt via two mechanisms: 
    1. a voltage dependent adaptation.  
    2. a component initiated by a spike which decays as an exponential.  
    These two component are in reference to threshold infinity and are recorded 
    in the neuron's threshold components.
    The third component refers to th_inf which is added separately as opposed to being 
    included in the voltage component of the threshold as is done in equation 2.1 of
    Mihalas and Nieber 2009.  Threshold infinity is removed for simple optimization.
    
    Parameters
    ----------
    neuron : class
    threshold_t0 : float
        threshold input to function
    voltage_t0 : float
        voltage input to function
    AScurrents_t0 : vector
        values of after spike currents 
    inj : float
        current injected into the neuron
    """
    tcs = neuron.get_threshold_components()

    th_spike = tcs['spike'][-1]
    th_voltage = tcs['voltage'][-1] 

    a_voltage = a_voltage * neuron.coeffs['a']
    b_voltage = b_voltage * neuron.coeffs['b']

    I = inj + np.sum(AScurrents_t0)
    C = neuron.C * neuron.coeffs['C']
    g = neuron.G * neuron.coeffs['G']

    voltage_component=voltage_component_of_threshold_exact(th_voltage, voltage_t0, I, neuron.dt, a_voltage, b_voltage, C, g, neuron.El)
    spike_component = spike_component_of_threshold_exact(th_spike, b_spike, neuron.dt)
 
    #------update the voltage and spiking values of the the
    neuron.add_threshold_components(spike_component, voltage_component)
     
    return voltage_component+spike_component+neuron.th_inf * neuron.coeffs['th_inf']    

def dynamics_threshold_spike_component(neuron, threshold_t0, voltage_t0, AScurrents_t0, inj,
                                       a_spike, b_spike, a_voltage, b_voltage):
    """Analytical solution for spike component of threshold. The threshold will adapt via 
    a component initiated by a spike which decays as an exponential.  The component is in 
    reference to threshold infinity and are recorded in the neuron's threshold components.  The voltage 
    component of the threshold is set to zero in the threshold components  because it is zero here  
    The third component refers to th_inf which is added separately as opposed to being 
    included in the voltage component of the threshold as is done in equation 2.1 of
    Mihalas and Nieber 2009.  Threshold infinity is removed for simple optimization.
    
    Parameters
    ----------
    neuron : class
    threshold_t0 : float
        threshold input to function
    voltage_t0 : float
        voltage input to function
    AScurrents_t0 : vector
        values of after spike currents 
    inj : float
        current injected into the neuron
    """

    tcs = neuron.get_threshold_components()

    th_spike = tcs['spike'][-1]
    th_voltage = tcs['voltage'][-1] 

    spike_component = spike_component_of_threshold_exact(th_spike, b_spike, neuron.dt)
 
    #------update the voltage and spiking values of the the
    neuron.add_threshold_components(spike_component, 0.0)
     
    return spike_component+neuron.th_inf * neuron.coeffs['th_inf']


def dynamics_threshold_inf(neuron, threshold_t0, voltage_t0, AScurrents_t0, inj):
    """ Set threshold to the neuron's instantaneous threshold. 

    Parameters
    ----------
    neuron : class
    threshold_t0 : not used here
    voltage_t0 : not used here
    AScurrents_t0 : not used here 
    inj : not used here
    AScurrents_t0 : not used here 
    inj : not used here
    """   
    return neuron.coeffs['th_inf'] * neuron.th_inf


def reset_AScurrent_sum(neuron, AScurrents_t0, r):
    """ Reset afterspike currents by adding summed exponentials. Left over currents from last spikes as 
    well as newly initiated currents from current spike.  Currents amplitudes in neuron.asc_amp_array need
    to be the amplitudes advanced though the spike cutting.  I.e. In the preprocessor if the after spike currents 
    are calculated via the GLM from spike initiation the amplitude at the time after the spike cutting needs to 
    be calculated and neuron.asc_amp_array needs to be set to this value.
    
    Parameters
    ----------
    r : np.ndarray
        a coefficient vector applied to the afterspike currents
    """
    new_currents=neuron.asc_amp_array * neuron.coeffs['asc_amp_array'] #neuron.asc_amp_array are amplitudes initiating after the spike is cut
    left_over_currents=AScurrents_t0 * r * np.exp(-(neuron.k * neuron.dt * neuron.spike_cut_length)) #advancing cut currents though the spike    

    return new_currents+left_over_currents


def reset_AScurrent_none(neuron, AScurrents_t0):
    """ Reset afterspike currents to zero. """
    
    if np.sum(AScurrents_t0)!=0:
        raise Exception('You are running a LIF but the AScurrents are not zero!')
    return np.zeros(len(AScurrents_t0))


def reset_voltage_v_before(neuron, voltage_t0, a, b):
    """ Reset voltage to the previous value with a scale and offset applied.

    Parameters
    ----------
    a : float
        voltage scale constant
    b : float
        voltage offset constant
    """

    return a*(voltage_t0)+b


def reset_voltage_i_v_before(neuron, voltage_t0):
    """ This method is not implemented yet.  It will raise an exception. """
    raise Exception("reset_voltage_IandVbefore not implemented")


def reset_voltage_zero(neuron, voltage_t0):
    """ Reset voltage to zero. """
    return 0.0


def reset_voltage_fixed(neuron, voltage_t0, value):
    """ Reset voltage to a fixed value. 

    Parameters
    ----------
    value : float
        the value to which voltage will be reset.
    """
    return value

def reset_threshold_max_v_th(neuron, threshold_t0, voltage_v1, delta):
    """ Return the maximum of threshold and reset voltage offset by a constant. 

    Parameters
    ----------
    delta : float
        value used to offset the return threshold.
    """
    #This is a bit dangerous as it would change if El was not choosen to be zero. Perhaps could change it to absolute value.
    return max(threshold_t0, voltage_v1) + delta  
    

def reset_threshold_th_before(neuron, threshold_t0, voltage_v1, delta):
    """ Return the previous threshold by a constant. This method is not used and will raise an exception if called.

    Parameters
    ----------
    delta : float
        value used to offset the return threshold.
    """
    raise Exception ('reset_threshold_th_before should not be called')
    return threshold_t0 + delta


def reset_threshold_inf(neuron, threshold_t0, voltage_v1):
    """ Reset the threshold to instantaneous threshold. """
    return neuron.coeffs['th_inf'] * neuron.th_inf


def reset_threshold_fixed(neuron, threshold_t0, voltage_v1, value):
    """ Reset the threshold to a fixed value. This method is not sued and will raise an exception if called.

    Parameters
    ----------
    value : float
        value to return as the reset threshold
    """
    raise Exception('reset_threshold_fixed should not be called')
    return  value


def reset_threshold_three_components(neuron, threshold_t0, voltage_v1, a_spike):
    '''This method resets the two components of the threshold: a spike (fast)
    component and a (voltage) component which are summed. 
    '''
    tcs = neuron.get_threshold_components()
    neuron.add_threshold_components( tcs['spike'][-1] + a_spike,  #adding spiking component of threshold ontop of already existent spike component.
                                     tcs['voltage'][-1] ) #note these are the same value.
    
    return tcs['spike'][-1] + tcs['voltage'][-1] + neuron.th_inf * neuron.coeffs['th_inf']


#: The METHOD_LIBRARY constant groups dynamics and reset methods by group name (e.g. 'voltage_dynamics_method').  
#Those groups assign each method in this file a string name.  This is used by the GlifNeuron when initializing 
#its dynamics and reset methods.
METHOD_LIBRARY = {
    'AScurrent_dynamics_method': { 
        'exp':    dynamics_AScurrent_exp,
        'vector': dynamics_AScurrent_vector,
        'none':   dynamics_AScurrent_none
        },
    'voltage_dynamics_method': { 
        'linear_forward_euler':           dynamics_voltage_linear_forward_euler,
        'quadratic_forward_euler':        dynamics_voltage_quadratic_forward_euler,
        'piecewise_linear_forward_euler': dynamics_voltage_piecewise_linear_forward_euler,
        'linear_exact':                   dynamics_voltage_linear_exact,
        'piecewise_linear_exact':         dynamics_voltage_piecewise_linear_exact,
        },
    'threshold_dynamics_method': {
        'spike_component':          dynamics_threshold_spike_component, 
        'inf':                      dynamics_threshold_inf,
        'three_components_forward': dynamics_threshold_three_components_forward, 
        'three_components_exact':   dynamics_threshold_three_components_exact 
        },
    'AScurrent_reset_method': {
        'sum':  reset_AScurrent_sum,
        'none': reset_AScurrent_none
        }, 
    'voltage_reset_method': {
        'v_before':   reset_voltage_v_before,
        'i_v_before': reset_voltage_i_v_before,
        'zero':       reset_voltage_zero,
        'fixed':      reset_voltage_fixed
        }, 
    'threshold_reset_method': {
        'max_v_th':         reset_threshold_max_v_th,
        'th_before':        reset_threshold_th_before,
        'inf':              reset_threshold_inf,
        'fixed':            reset_threshold_fixed,
        'three_components': reset_threshold_three_components
        }
}



# Copyright 2015-2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.


import pytest
from mock import patch, mock_open
from allensdk.core.json_utilities import JsonComments


@pytest.fixture
def commented_json():
    return ("{\n"
            "    // comment\n"
            "    \"color\": \"blue\"\n"
            "}")


@pytest.fixture
def blank_line_json():
    return ("{\n"
            "\n"
            "\n"
            "\n"
            "    \"color\": \"blue\"\n"
            "}")


@pytest.fixture
def multi_line_json():
    return ("{\n"
            "/* \n"
            " * multiline comment\n"
            " */\n"
            "    \"color\": \"blue\"\n"
            "}")


@pytest.fixture
def two_multi_line_json():
    return ("{\n"
            "    \"colors\": [\"blue\",\n"
            "    /* comment these out\n"
            "    \"red\",\n"
            "    \"yellow\",\n"
            "     ... but not these */\n"
            "    \"orange\",\n"
            "    \"purple\",\n"
            "    /* also comment this out\n"
            "    \"indigo\",\n"
            "    .... end comment */\n"
            "    \"violet\"\n"
            "    ]\n"
            "}")


def testSingleLineComment(commented_json):
    parsed_json = JsonComments.read_string(
        commented_json)
    
    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')


def testBlankLines(blank_line_json):
    parsed_json = JsonComments.read_string(
        blank_line_json)
    
    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')


def testMultiLineComment(multi_line_json):
    parsed_json = JsonComments.read_string(
        multi_line_json)
    
    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')


def testTwoMultiLineComments(two_multi_line_json):
    parsed_json = JsonComments.read_string(
        two_multi_line_json)
    
    assert('colors' in parsed_json)
    assert(len(parsed_json['colors']) == 4)
    assert('blue' in parsed_json['colors'])
    assert('orange' in parsed_json['colors'])
    assert('purple' in parsed_json['colors'])
    assert('violet' in parsed_json['colors'])


def testSingleLineCommentFile(commented_json):
    with patch("__builtin__.open",
               mock_open(
                   read_data=commented_json)):
        parsed_json = JsonComments.read_file('mock.json')
     
    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')
 
 
def testBlankLinesFile(blank_line_json):
    with patch("__builtin__.open",
               mock_open(
                   read_data=blank_line_json)):
        parsed_json = JsonComments.read_file('mock.json')
     
    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')
 
 
def testMultiLineFile(multi_line_json):
    with patch("__builtin__.open",
               mock_open(
                   read_data=multi_line_json)):
        parsed_json = JsonComments.read_file('mock.json')
     
    assert('color' in parsed_json and
           parsed_json['color'] == 'blue')
 
 
def testTwoMultiLineFile(two_multi_line_json):
    with patch("__builtin__.open",
               mock_open(
                   read_data=two_multi_line_json)):
        parsed_json = JsonComments.read_file('mock.json')
     
    assert('colors' in parsed_json)
    assert(len(parsed_json['colors']) == 4)
    assert('blue' in parsed_json['colors'])
    assert('orange' in parsed_json['colors'])
    assert('purple' in parsed_json['colors'])
    assert('violet' in parsed_json['colors'])



# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.


import pytest
from allensdk.config.manifest_builder import ManifestBuilder


@pytest.fixture
def builder():
    b = ManifestBuilder()
    b.add_path('BASEDIR', '/home/username/example')
    
    return b


def testManifestConstructor(builder):
    manifest = builder.get_manifest()
    expected = '/home/username/example'
    actual = manifest.get_path('BASEDIR')
    assert(expected == actual)


def testManifestParent(builder):
    builder.add_path('WORKDIR',
                     'work',
                     parent_key='BASEDIR')
    manifest = builder.get_manifest()
    expected = '/home/username/example/work'
    actual = manifest.get_path('WORKDIR')
    assert(expected == actual)


def testManifestBuilderDataFrame(builder):
    builder.add_path('WORKDIR',
                     'work',
                     parent_key='BASEDIR')
    builder_df = builder.as_dataframe()
    
    assert('key' in builder_df.keys())
    assert('type' in builder_df.keys())
    assert('spec' in builder_df.keys())
    assert('parent_key' in builder_df.keys())
    assert('format' in builder_df.keys())        
    assert(5 == len(builder_df.keys()))


def testManifestDataFrame(builder):
    builder.add_path('WORKDIR',
                     'work',
                     parent_key='BASEDIR')
    
    manifest = builder.get_manifest()
    df = manifest.as_dataframe()
    
    assert('type' in df.keys())
    assert('spec' in df.keys())
    assert(2 == len(df.keys()))



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import os
import pandas as pd

from allensdk.config.manifest_builder import ManifestBuilder
from allensdk.api.cache import Cache
from allensdk.api.queries.cell_types_api import CellTypesApi

import allensdk.core.json_utilities as json_utilities
from allensdk.core.nwb_data_set import NwbDataSet
import allensdk.core.swc as swc

import logging

class CellTypesCache(Cache):
    """
    Cache class for storing and accessing data from the Cell Types Database.
    By default, this class will cache any downloaded metadata or files in 
    well known locations defined in a manifest file.  This behavior can be
    disabled.

    Attributes
    ----------
    
    api: CellTypesApi instance
        The object used for making API queries related to the Cell Types Database

    Parameters
    ----------
    
    cache: boolean
        Whether the class should save results of API queries to locations specified
        in the manifest file.  Queries for files (as opposed to metadata) must have a
        file location.  If caching is disabled, those locations must be specified
        in the function call (e.g. get_ephys_data(file_name='file.nwb')).

    manifest_file: string
       File name of the manifest to be read.  Default is "cell_types_manifest.json".
    """

    # manifest keys
    CELLS_KEY = 'CELLS'
    EPHYS_FEATURES_KEY = 'EPHYS_FEATURES'
    MORPHOLOGY_FEATURES_KEY = 'MORPHOLOGY_FEATURES'
    EPHYS_DATA_KEY = 'EPHYS_DATA'
    EPHYS_SWEEPS_KEY = 'EPHYS_SWEEPS'
    RECONSTRUCTION_KEY = 'RECONSTRUCTION'
    MARKER_KEY = 'MARKER'

    def __init__(self, cache=True, manifest_file='cell_types_manifest.json', base_uri=None):
        super(CellTypesCache, self).__init__(manifest=manifest_file, cache=cache)
        self.api = CellTypesApi(base_uri=base_uri)


    def get_cells(self, file_name=None, require_morphology=False, require_reconstruction=False, reporter_status=None):
        """
        Download metadata for all cells in the database and optionally return a
        subset filtered by whether or not they have a morphology or reconstruction.

        Parameters
        ----------
        
        file_name: string
            File name to save/read the cell metadata as JSON.  If file_name is None, 
            the file_name will be pulled out of the manifest.  If caching
            is disabled, no file will be saved. Default is None.

        require_morphology: boolean
            Filter out cells that have no morphological images.

        require_reconstruction: boolean
            Filter out cells that have no morphological reconstructions.
           
        reporter_status: list
            Filter for cells that have one or more cell reporter statuses.
        """

        file_name = self.get_cache_path(file_name, self.CELLS_KEY)

        if os.path.exists(file_name):
            cells = json_utilities.read(file_name)
        else:
            cells = self.api.list_cells(False, False)

            if self.cache:
                json_utilities.write(file_name, cells)

        if isinstance(reporter_status, basestring):
            reporter_status = [ reporter_status ]

        # filter the cells on the way out
        return self.api.filter_cells(cells, require_morphology, require_reconstruction, reporter_status)
    

    def get_ephys_sweeps(self, specimen_id, file_name=None):
        """
        Download sweep metadata for a single cell specimen.  

        Parameters
        ----------
        
        specimen_id: int
             ID of a cell.
        """

        file_name = self.get_cache_path(file_name, self.EPHYS_SWEEPS_KEY, specimen_id)
        
        if os.path.exists(file_name):
            sweeps = json_utilities.read(file_name)
        else:
            sweeps = self.api.get_ephys_sweeps(specimen_id)

            if self.cache:
                json_utilities.write(file_name, sweeps)

        return sweeps

    
    def get_ephys_features(self, dataframe=False, file_name=None):
        """
        Download electrophysiology features for all cells in the database.

        Parameters
        ----------
        
        file_name: string
            File name to save/read the ephys features metadata as CSV.  
            If file_name is None, the file_name will be pulled out of the 
            manifest.  If caching is disabled, no file will be saved. 
            Default is None.

        dataframe: boolean
            Return the output as a Pandas DataFrame.  If False, return 
            a list of dictionaries.
        """

        file_name = self.get_cache_path(file_name, self.EPHYS_FEATURES_KEY)

        if os.path.exists(file_name):
            features_df = pd.DataFrame.from_csv(file_name)
        else:
            features_df = self.api.get_ephys_features(dataframe=True)

            if self.cache:
                features_df.to_csv(file_name)

        if dataframe:
            return features_df
        else:
            return features_df.to_dict('records')


    def get_morphology_features(self, dataframe=False, file_name=None):
        """
        Download morphology features for all cells with reconstructions in the database.

        Parameters
        ----------
        
        file_name: string
            File name to save/read the ephys features metadata as CSV.  
            If file_name is None, the file_name will be pulled out of the 
            manifest.  If caching is disabled, no file will be saved. 
            Default is None.

        dataframe: boolean
            Return the output as a Pandas DataFrame.  If False, return 
            a list of dictionaries.
        """
        
        file_name = self.get_cache_path(file_name, self.MORPHOLOGY_FEATURES_KEY)
        
        if os.path.exists(file_name):
            features_df = pd.DataFrame.from_csv(file_name)
        else:
            features_df = self.api.get_morphology_features(dataframe=True)
            
            if self.cache:
                features_df.to_csv(file_name)

        if dataframe:
            return features_df
        else:
            return features_df.to_dict('records')

    
    def get_all_features(self, dataframe=False, require_reconstruction=True):
        """
        Download morphology and electrophysiology features for all cells and merge them
        into a single table.

        Parameters
        ----------

        dataframe: boolean
            Return the output as a Pandas DataFrame.  If False, return 
            a list of dictionaries.

        require_reconstruction: boolean
            Only return ephys and morphology features for cells that have 
            reconstructions. Default True.
        """

        ephys_features = self.get_ephys_features(dataframe=True)
        morphology_features = self.get_morphology_features(dataframe=True)

        how = 'inner' if require_reconstruction else 'outer'

        all_features = ephys_features.merge(morphology_features, 
                                            how=how,
                                            on='specimen_id')
        
        if dataframe:
            return all_features
        else:
            return all_features.to_dict('records')
        

    def get_ephys_data(self, specimen_id, file_name=None):
        """
        Download electrophysiology traces for a single cell in the database.

        Parameters
        ----------
        
        specimen_id: int
            The ID of a cell specimen to download.

        file_name: string
            File name to save/read the ephys features metadata as CSV.  
            If file_name is None, the file_name will be pulled out of the 
            manifest.  If caching is disabled, no file will be saved. 
            Default is None.

        Returns
        -------
        NwbDataSet
            A class instance with helper methods for retrieving stimulus
            and response traces out of an NWB file.
        """

        file_name = self.get_cache_path(file_name, self.EPHYS_DATA_KEY, specimen_id)

        if not os.path.exists(file_name):
            self.api.save_ephys_data(specimen_id, file_name)

        return NwbDataSet(file_name)


    def get_reconstruction(self, specimen_id, file_name=None):
        """
        Download and open a reconstruction for a single cell in the database.

        Parameters
        ----------
        
        specimen_id: int
            The ID of a cell specimen to download.

        file_name: string
            File name to save/read the reconstruction SWC.  
            If file_name is None, the file_name will be pulled out of the 
            manifest.  If caching is disabled, no file will be saved. 
            Default is None.

        Returns
        -------
        Morphology
             A class instance with methods for accessing morphology compartments.
        """

        file_name = self.get_cache_path(file_name, self.RECONSTRUCTION_KEY, specimen_id)

        if file_name is None:
            raise Exception("Please enable caching (CellTypes.cache = True) or specify a save_file_name.")

        if not os.path.exists(file_name):
            self.api.save_reconstruction(specimen_id, file_name)

        return swc.read_swc(file_name)


    def get_reconstruction_markers(self, specimen_id, file_name=None):
        """
        Download and open a reconstruction marker file for a single cell in the database.

        Parameters
        ----------
        
        specimen_id: int
            The ID of a cell specimen to download.

        file_name: string
            File name to save/read the reconstruction marker.  
            If file_name is None, the file_name will be pulled out of the 
            manifest.  If caching is disabled, no file will be saved. 
            Default is None.

        Returns
        -------
        Morphology
             A class instance with methods for accessing morphology compartments.
        """

        file_name = self.get_cache_path(file_name, self.MARKER_KEY, specimen_id)

        if file_name is None:
            raise Exception("Please enable caching (CellTypes.cache = True) or specify a save_file_name.")

        if not os.path.exists(file_name):
            try:
                self.api.save_reconstruction_markers(specimen_id, file_name)
            except LookupError as e:
                logging.warning(e.message)
                return []

        return swc.read_marker_file(file_name)


    def build_manifest(self, file_name):
        """
        Construct a manifest for this Cache class and save it in a file.
        
        Parameters
        ----------
        
        file_name: string
            File location to save the manifest.

        """

        mb = ManifestBuilder()

        mb.add_path('BASEDIR', '.')
        mb.add_path(self.CELLS_KEY, 'cells.json', typename='file', parent_key='BASEDIR')
        mb.add_path(self.EPHYS_DATA_KEY, 'specimen_%d/ephys.nwb', typename='file', parent_key='BASEDIR')
        mb.add_path(self.EPHYS_FEATURES_KEY, 'ephys_features.csv', typename='file', parent_key='BASEDIR')
        mb.add_path(self.MORPHOLOGY_FEATURES_KEY, 'morphology_features.csv', typename='file', parent_key='BASEDIR')
        mb.add_path(self.RECONSTRUCTION_KEY, 'specimen_%d/reconstruction.swc', typename='file', parent_key='BASEDIR')
        mb.add_path(self.MARKER_KEY, 'specimen_%d/reconstruction.marker', typename='file', parent_key='BASEDIR')
        mb.add_path(self.EPHYS_SWEEPS_KEY, 'specimen_%d/ephys_sweeps.json', typename='file', parent_key='BASEDIR')


        mb.write_json_file(file_name)


class ReporterStatus:
    """
    Valid strings for filtering by cell reporter status.
    """

    POSITIVE = 'cre reporter positive'
    NEGATIVE = 'cre reporter negative'
    NA = 'not applicable'
    INDETERMINATE = 'cre reporter indeterminate'




    



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from allensdk.api.api import Api

class TreeSearchApi(Api):
    '''
    
    See `Searching a Specimen or Structure Tree <http://help.brain-map.org/display/api/Image-to-Image+Synchronization>`_
    for additional documentation.
    '''
    
    def __init__(self, base_uri=None):
        super(TreeSearchApi, self).__init__(base_uri)    
        
    
    def get_tree(self,
                 kind,
                 db_id,
                 ancestors=None,
                 descendants=None):
        '''Fetch meta data for the specified structure or specimen.
        
        Parameters
        ----------
        kind : string
            'Structure' or 'Specimen'
        db_id : integer
            The id of the structure or specimen to search.
        ancestors : boolean, optional
            whether to include ancestors in the response (defaults to False)
        descendants : boolean, optional
            whether to include descendants in the response (defaults to False)
        
        Returns
        -------
        dict
            parsed json response data
        '''
        params = []
        url_params = ''
        
        if ancestors == True:
            params.append('ancestors=true')
        elif ancestors == False:
            params.append('ancestors=false')
        
        if descendants == True:
            params.append('descendants=true')
        elif descendants == False:
            params.append('descendants=false')
        
        if len(params) > 0:
            url_params = '?' + '&'.join(params)
        else:
            url_params = ''
        
        url = ''.join([self.tree_search_endpoint,
                       '/',
                       kind,
                       '/',
                       str(db_id),
                       '.json',
                       url_params])
        
        return self.json_msg_query(url)


# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from allensdk.config.manifest import Manifest
import allensdk.core.json_utilities as ju
import pandas as pd
import pandas.io.json as pj
import os


class Cache(object):
    def __init__(self,
                 manifest=None,
                 cache=True):
        self.cache = cache
        self.load_manifest(manifest)
    

    def get_cache_path(self, file_name, manifest_key, *args):
        '''Helper method for accessing path specs from manifest keys.
        
        Parameters
        ----------
        file_name : string
        manifest_key : string
        args : ordered parameters
    
        Returns
        -------
        string or None
            path
        '''
        if self.cache:
            if file_name:
                return file_name
            elif self.manifest:
                return self.manifest.get_path(manifest_key, *args)

        return None
        

    def load_manifest(self, file_name):
        '''Read a keyed collection of path specifications.
        
        Parameters
        ----------
        file_name : string
            path to the manifest file
        
        Returns
        -------
        Manifest
        '''
        if file_name != None:
            if not os.path.exists(file_name):

                # make the directory if it doesn't exist already
                dirname = os.path.dirname(file_name)
                Manifest.safe_mkdir(dirname)

                self.build_manifest(file_name)

            
            self.manifest = Manifest(ju.read(file_name)['manifest'], os.path.dirname(file_name))
        else:
            self.manifest = None


    def build_manifest(self, file_name):
        '''Creation of default path speifications.
        
        Parameters
        ----------
        file_name : string
            where to save it
        '''
        raise Exception("This function must be defined in the appropriate subclass")


    def manifest_dataframe(self):
        '''Convenience method to view manifest as a pandas dataframe.
        '''
        return pd.DataFrame.from_dict(self.manifest.path_info,
                                      orient='index')

    def rename_columns(self,
                       data,
                       new_old_name_tuples=None):
        '''Convenience method to rename columns in a pandas dataframe.
        
        Parameters
        ----------
        data : dataframe
            edited in place.
        new_old_name_tuples : list of string tuples (new, old)
        '''
        if new_old_name_tuples == None:
            new_old_name_tuples = []
            
        for new_name, old_name in new_old_name_tuples:
            data.columns = [new_name if c == old_name else c
                            for c in data.columns]                    
    
    
    def load_csv(self,
                 path,
                 rename=None,
                 index=None):
        '''Read a csv file as a pandas dataframe.
        
        Parameters
        ----------
        rename : list of string tuples (new old), optional
            columns to rename
        index : string, optional
            post-rename column to use as the row label.
        '''
        data = pd.DataFrame.from_csv(path)

        self.rename_columns(data, rename)

        if index is not None:        
            data.set_index([index], inplace=True)

        return data


    def load_json(self,
                  path,
                  rename=None,
                  index=None):
        '''Read a json file as a pandas dataframe.
        
        Parameters
        ----------
        rename : list of string tuples (new old), optional
            columns to rename
        index : string, optional
            post-rename column to use as the row label.
        '''
        data = pj.read_json(path, orient='records')

        self.rename_columns(data, rename)
        
        if index is not None:        
            data.set_index([index], inplace=True)

        return data
    

    def wrap(self, fn, path, cache,
             save_as_json=True,
             return_dataframe=False,
             index=None,
             rename=None,
             **kwargs):
        '''make an rma query, save it and return the dataframe.
        
        Parameters
        ----------
        fn : function reference
            makes the actual query using kwargs.
        path : string
            where to save the data
        cache : boolean
            True will make the query, False just loads from disk
        save_as_json : boolean, optional
            True (default) will save data as json, False as csv
        return_dataframe : boolean, optional
            True will cast the return value to a pandas dataframe, False (default) will not 
        index : string, optional
            column to use as the pandas index
        rename : list of string tuples, optional
            (new, old) columns to rename
        kwargs : objects
            passed through to the query function
        
        Returns
        -------
        dict or DataFrame
            data type depends on return_dataframe option.
        
        Notes
        -----
        Column renaming happens after the file is reloaded for json
        '''
        if cache == True:
            json_data = fn(**kwargs)
            
            if save_as_json == True:
                ju.write(path, json_data)
            else:
                df = pd.DataFrame(json_data)
                self.rename_columns(df, rename)
                
                if index is not None:
                    df.set_index([index], inplace=True)
        
                df.to_csv(path)
    
        # read it back in
        if save_as_json == True:
            if return_dataframe == True:
                data = pj.read_json(path, orient='records')
                self.rename_columns(data, rename)
                if index != None:
                    data.set_index([index], inplace=True)
            else:
                data = ju.read(path)
        elif return_dataframe == True:
            data = pd.DataFrame.from_csv(path)
        else:
            raise ValueError('save_as_json=False cannot be used with return_dataframe=False')
        
        return data



# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from allensdk.brain_observatory.static_gratings import StaticGratings
from allensdk.brain_observatory.locally_sparse_noise import LocallySparseNoise
from allensdk.brain_observatory.natural_scenes import NaturalScenes
from allensdk.brain_observatory.drifting_gratings import DriftingGratings
from allensdk.brain_observatory.natural_movie import NaturalMovie

from allensdk.core.brain_observatory_nwb_data_set import BrainObservatoryNwbDataSet
import allensdk.brain_observatory.stimulus_info as stimulus_info
from allensdk.brain_observatory.brain_observatory_exceptions import BrainObservatoryAnalysisException
import allensdk.brain_observatory.brain_observatory_plotting as cp
import argparse, logging, os
import sys
import numpy as np

def multi_dataframe_merge(dfs):
    out_df = None
    for i,df in enumerate(dfs):
        if out_df is None:
            out_df = df
        else:
            out_df = out_df.merge(df, left_index=True, right_index=True, suffixes=['','_deleteme'])

    bad_columns = set([ c for c in out_df.columns if c.endswith('deleteme') ])
    out_df.drop(list(bad_columns), axis=1, inplace=True)

    return out_df

class SessionAnalysis(object):
    _log = logging.getLogger('allensdk.brain_observatory.session_analysis')    

    def __init__(self, nwb_path, save_path):
        self.nwb = BrainObservatoryNwbDataSet(nwb_path)                        
        self.save_path = save_path
        self.save_dir = os.path.dirname(save_path)

        self.metrics_a = {}
        self.metrics_b = {}
        self.metrics_c = {}

        self.metadata = self.nwb.get_metadata()

    def append_metadata(self, df):
        for k,v in self.metadata.iteritems():
            df[k] = v

    def save_session_a(self, dg, nm1, nm3, peak):
        nwb = BrainObservatoryNwbDataSet(self.save_path)
        nwb.save_analysis_dataframes(
            ('stim_table_dg', dg.stim_table),
            ('sweep_response_dg', dg.sweep_response),
            ('mean_sweep_response_dg', dg.mean_sweep_response),
            ('peak', peak),        
            ('sweep_response_nm1', nm1.sweep_response),
            ('stim_table_nm1', nm1.stim_table),
            ('sweep_response_nm3', nm3.sweep_response))
        
        nwb.save_analysis_arrays(
            ('celltraces_dff', nm1.dfftraces),
            ('response_dg', dg.response),
            ('binned_cells_sp', nm1.binned_cells_sp),
            ('binned_cells_vis', nm1.binned_cells_vis),
            ('binned_dx_sp', nm1.binned_dx_sp),
            ('binned_dx_vis', nm1.binned_dx_vis))
    
        
    def save_session_b(self, sg, nm1, ns, peak): 
        nwb = BrainObservatoryNwbDataSet(self.save_path)
        nwb.save_analysis_dataframes(
            ('stim_table_sg', sg.stim_table),
            ('sweep_response_sg', sg.sweep_response),
            ('mean_sweep_response_sg', sg.mean_sweep_response),
            ('sweep_response_nm1', nm1.sweep_response),
            ('stim_table_nm1', nm1.stim_table),
            ('sweep_response_ns', ns.sweep_response),
            ('stim_table_ns', ns.stim_table),
            ('mean_sweep_response_ns', ns.mean_sweep_response),
            ('peak', peak))

        nwb.save_analysis_arrays(
            ('celltraces_dff', nm1.dfftraces),
            ('response_sg', sg.response),
            ('response_ns', ns.response),
            ('binned_cells_sp', nm1.binned_cells_sp),
            ('binned_cells_vis', nm1.binned_cells_vis),
            ('binned_dx_sp', nm1.binned_dx_sp),
            ('binned_dx_vis', nm1.binned_dx_vis))
    
    
    def save_session_c(self, lsn, nm1, nm2, peak):                
        nwb = BrainObservatoryNwbDataSet(self.save_path)
        nwb.save_analysis_dataframes(
            ('stim_table_lsn', lsn.stim_table),
            ('sweep_response_nm1', nm1.sweep_response),
            ('peak', peak),
            ('sweep_response_nm2', nm2.sweep_response),
            ('sweep_response_lsn', lsn.sweep_response),
            ('mean_sweep_response_lsn', lsn.mean_sweep_response))  
        
        nwb.save_analysis_arrays(
            ('receptive_field_lsn', lsn.receptive_field),
            ('celltraces_dff', nm1.dfftraces),
            ('binned_dx_sp', nm1.binned_dx_sp),
            ('binned_dx_vis', nm1.binned_dx_vis),    
            ('binned_cells_sp', nm1.binned_cells_sp),
            ('binned_cells_vis', nm1.binned_cells_vis))
    
    def append_metrics_drifting_grating(self, metrics, dg):
        metrics["osi_dg"] = dg.peak["osi_dg"]
        metrics["dsi_dg"] = dg.peak["dsi_dg"]
        metrics["pref_dir_dg"] = [ dg.orivals[i] for i in dg.peak["ori_dg"].values ]
        metrics["pref_tf_dg"] = [ dg.tfvals[i] for i in dg.peak["tf_dg"].values ]
        metrics["p_dg"] = dg.peak["ptest_dg"]
    
    def append_metrics_static_grating(self, metrics, sg):
        metrics["osi_sg"] = sg.peak["osi_sg"]
        metrics["pref_ori_sg"] = [ sg.orivals[i] for i in sg.peak["ori_sg"].values ]
        metrics["pref_sf_sg"] = [ sg.sfvals[i] for i in sg.peak["sf_sg"].values ]
        metrics["pref_phase_sg"] = [ sg.phasevals[i] for i in sg.peak["phase_sg"].values ]
        metrics["p_sg"] = sg.peak["ptest_sg"]
        metrics["time_to_peak_sg"] = sg.peak["time_to_peak_sg"]

    def append_metrics_natural_scene(self, metrics, ns):
        metrics["pref_image_ns"] = ns.peak["scene_ns"]
        metrics["p_ns"] = ns.peak["ptest_ns"]
        metrics["time_to_peak_ns"] = ns.peak["time_to_peak_ns"]

    def verify_roi_lists_equal(self, roi1, roi2):
        if len(roi1) != len(roi2):
            raise BrainObservatoryAnalysisException("Error -- ROI lists are of different length")
        for i in range(len(roi1)):
            if roi1[i] != roi2[i]:
                raise BrainObservatoryAnalysisException("Error -- ROI lists have different entries")
    
    def session_a(self, plot_flag=False, save_flag=True):
        nm1 = NaturalMovie(self.nwb, 'natural_movie_one', speed_tuning=True)
        nm3 = NaturalMovie(self.nwb, 'natural_movie_three')
        dg = DriftingGratings(self.nwb)

        SessionAnalysis._log.info("Session A analyzed")
        peak = multi_dataframe_merge([nm1.peak_run, dg.peak, nm1.peak, nm3.peak])
        

        self.append_metrics_drifting_grating(self.metrics_a, dg)
        self.metrics_a["roi_id"] = dg.roi_id

        self.append_metadata(peak)

        if save_flag:
            self.save_session_a(dg, nm1, nm3, peak)

        if plot_flag:
            cp._plot_3sa(dg, nm1, nm3, self.save_dir)
            cp.plot_drifting_grating_traces(dg, self.save_dir)
    
    
    def session_b(self, plot_flag=False, save_flag=True):
        ns = NaturalScenes(self.nwb)
        sg = StaticGratings(self.nwb)
        nm1 = NaturalMovie(self.nwb, 'natural_movie_one', speed_tuning=True)
        SessionAnalysis._log.info("Session B analyzed")
        peak = multi_dataframe_merge([nm1.peak_run, sg.peak, ns.peak, nm1.peak])
        self.append_metadata(peak)

        self.append_metrics_static_grating(self.metrics_b, sg)
        self.append_metrics_natural_scene(self.metrics_b, ns)
        self.verify_roi_lists_equal(sg.roi_id, ns.roi_id)
        self.metrics_b["roi_id"] = sg.roi_id
            
        if save_flag:
            self.save_session_b(sg, nm1, ns, peak)    

        if plot_flag:
            cp._plot_3sb(sg, nm1, ns, self.save_dir)
            cp.plot_ns_traces(ns, self.save_dir)
            cp.plot_sg_traces(sg, self.save_dir)
                    
    
    def session_c(self, plot_flag=False, save_flag=True):
        lsn = LocallySparseNoise(self.nwb)
        nm2 = NaturalMovie(self.nwb, 'natural_movie_two')
        nm1 = NaturalMovie(self.nwb, 'natural_movie_one', speed_tuning=True)
        SessionAnalysis._log.info("Session C analyzed")
        peak = multi_dataframe_merge([nm1.peak_run, nm1.peak, nm2.peak])
        self.append_metadata(peak)
                
        self.metrics_c["roi_id"] = nm1.roi_id
        
        if save_flag:
            self.save_session_c(lsn, nm1, nm2, peak)

        if plot_flag:
            cp._plot_3sc(lsn, nm1, nm2, self.save_dir)
            cp.plot_lsn_traces(lsn, self.save_dir)
    
                    
def run_session_analysis(nwb_path, save_path, plot_flag=False):
    save_dir = os.path.abspath(os.path.dirname(save_path))

    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    session_analysis = SessionAnalysis(nwb_path, save_path)

    session = session_analysis.nwb.get_session_type()

    if session == stimulus_info.THREE_SESSION_A:
        session_analysis.session_a(plot_flag)
        metrics = session_analysis.metrics_a
    elif session == stimulus_info.THREE_SESSION_B:
        session_analysis.session_b(plot_flag)
        metrics = session_analysis.metrics_b
    elif session == stimulus_info.THREE_SESSION_C:
        session_analysis.session_c(plot_flag)
        metrics = session_analysis.metrics_c
    else:
        raise IndexError("Unknown session: %s" % session)

    return metrics
    
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_nwb")
    parser.add_argument("output_h5")

    parser.add_argument("--plot", action='store_true')

    args = parser.parse_args()
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)

    run_session_analysis(args.input_nwb, args.output_h5, args.plot)


if __name__=='__main__': main()
    



from allensdk.api.queries.glif_api import GlifApi
import allensdk.core.json_utilities as ju
from find_optimization_sweeps import find_optimization_sweeps

def test_find_optimization_sweeps():
    ga = GlifApi()
    nm = ga.get_neuronal_model(473836744)
    sweeps = ga.get_ephys_sweeps()
    
    opt_sweeps, stim_index, errs = find_optimization_sweeps(sweeps)



# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import os, unittest

from allensdk.core.nwb_data_set import NwbDataSet
from allensdk.core.cell_types_cache import CellTypesCache

class NwbDataSetTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(NwbDataSetTest, self).__init__(*args, **kwargs)

    def testAllDataSets(self):

        manifest_file = '/local1/projects/FHL2015/cell_types/manifest.json'
        if not os.path.exists(manifest_file):
            print "Cannot run this test: manifest does not exist (%s)" % manifest_file
            return True
        
        self.cache = CellTypesCache(manifest_file=manifest_file)
        cells = self.cache.get_cells()

        for cell in cells:
            data_set = self.cache.get_ephys_data(cell['id'])
            sweeps = self.cache.get_ephys_sweeps(cell['id'])

            for sweep in sweeps:
                metadata = data_set.get_sweep_metadata(sweep['sweep_number'])

if __name__ == "__main__":
    unittest.main()






import os, re, logging
import allensdk.core.json_utilities as ju

from optimize_neuron import optimize_neuron

DATA_CONFIG_PATTERN = "/data/mat/Corinne/GLIF_subset/data_config_files/%d_data_config.json"

MODEL_CONFIG_DIR = "/local1/stash/allensdk/allensdk/model/glif/test/model_config/"
MODEL_CONFIG_FILES = [ os.path.join(MODEL_CONFIG_DIR, "329552531_LIF_model_config.json"),                      
                       os.path.join(MODEL_CONFIG_DIR, "476106176_LIF_R_ASC_AT_PWL_model_config.json"),
                       os.path.join(MODEL_CONFIG_DIR, "329552531_LIF_ASC_model_config.json") ]
                       
OUT_DIR = "/local1/stash/allensdk/allensdk/model/glif/test/optimized_config/"

def test_optimize_neuron():
    p = re.compile("(\d+)_(.*)_model_config.json")

    if not os.path.exists(OUT_DIR):
        os.makedirs(OUT_DIR)

    for model_config_file in MODEL_CONFIG_FILES:
        logging.info("testing %s" % model_config_file)

        fname = os.path.basename(model_config_file)
        m = p.match(fname)

        sid, config = m.groups()
        data_config_file = DATA_CONFIG_PATTERN % int(sid)

        data_config = ju.read(data_config_file)
        nwb_file = data_config["filename"]
        sweep_list = data_config["sweeps"].values()

        model_config = ju.read(model_config_file)

        #DBG
        model_config['optimizer']['inner_iterations'] = 1
        model_config['optimizer']['outer_iterations'] = 1
        #DBG
        
        sweep_index = { s['sweep_number']:s for s in sweep_list }    

        optimizer, best_param, begin_param = optimize_neuron(model_config, sweep_index)

        out_file = os.path.join(OUT_DIR, "%s_%s_neuron_config.json" % (sid, config))
        ju.write(out_file, optimizer.experiment.neuron.to_dict())

        out_config_file = os.path.join(OUT_DIR, "%s_%s_optimized_model_config.json" % (sid, config))
        ju.write(out_config_file, {
                'optimizer': optimizer.to_dict(),
                'neuron': optimizer.experiment.neuron.to_dict()
                })


if __name__ == "__main__": test_optimize_neuron()



# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.


import scipy.stats as st
import numpy as np
import pandas as pd
from math import sqrt
import logging
from allensdk.brain_observatory.stimulus_analysis import StimulusAnalysis
from allensdk.brain_observatory.brain_observatory_exceptions \
    import BrainObservatoryAnalysisException


class StaticGratings(StimulusAnalysis):
    """ Perform tuning analysis specific to static gratings stimulus. 
    
    Parameters
    ----------
    data_set: BrainObservatoryNwbDataSet object
    """

    _log = logging.getLogger('allensdk.brain_observatory.static_gratings')        
    
    def __init__(self, data_set, **kwargs):
        super(StaticGratings, self).__init__(data_set, **kwargs)
        
        stimulus_table = self.data_set.get_stimulus_table('static_gratings')
        self.stim_table = stimulus_table.fillna(value=0.)     
        self.sweeplength = self.stim_table['end'].iloc[1] - self.stim_table['start'].iloc[1]
        self.interlength = 4 * self.sweeplength
        self.extralength = self.sweeplength
        self.orivals = np.unique(self.stim_table.orientation.dropna())
        self.sfvals = np.unique(self.stim_table.spatial_frequency.dropna())
        self.phasevals = np.unique(self.stim_table.phase.dropna())
        self.number_ori = len(self.orivals)
        self.number_sf = len(self.sfvals)
        self.number_phase = len(self.phasevals)            
        self.sweep_response, self.mean_sweep_response, self.pval = self.get_sweep_response()
        self.response = self.get_response()
        self.peak = self.get_peak()
        
    def get_response(self):
        ''' Computes the mean response for each cell to each stimulus condition.  Return is 
        a (# orientations, # spatial frequencies, # phasees, # cells, 3) np.ndarray.  The final dimension
        contains the mean response to the condition (index 0), standard error of the mean of the response
        to the condition (index 1), and p value of the response to that condition (index 3).

        Returns
        -------
        Numpy array storing mean responses.
        '''
        StaticGratings._log.info("Calculating mean responses")
        
        response = np.empty((self.number_ori, self.number_sf, self.number_phase, self.numbercells+1, 3))

        
        def ptest(x):
            return len(np.where(x<(0.05/(self.number_ori*(self.number_sf-1))))[0])
            
        for ori in self.orivals:
            ori_pt = np.where(self.orivals == ori)[0][0]
            
            for sf in self.sfvals:
                sf_pt = np.where(self.sfvals == sf)[0][0]
                
                for phase in self.phasevals:
                    phase_pt = np.where(self.phasevals == phase)[0][0]
                    subset_response = self.mean_sweep_response[(self.stim_table.spatial_frequency==sf)&(self.stim_table.orientation==ori)&(self.stim_table.phase==phase)]
                    subset_pval = self.pval[(self.stim_table.spatial_frequency==sf)&(self.stim_table.orientation==ori)&(self.stim_table.phase==phase)]
                    response[ori_pt, sf_pt, phase_pt, :, 0] = subset_response.mean(axis=0)
                    response[ori_pt, sf_pt, phase_pt, :, 1] = subset_response.std(axis=0)/sqrt(len(subset_response))
                    response[ori_pt, sf_pt, phase_pt, :, 2] = subset_pval.apply(ptest, axis=0)

        return response
    
    
    def get_peak(self):    
        ''' Computes metrics related to each cell's peak response condition.
        
        Returns
        -------
        Panda data frame with the following fields (_sg suffix is 
        for static grating):
            * ori_sg (orientation)
            * sf_sg (spatial frequency)
            * phase_sg 
            * response_variability_sg
            * osi_sg (orientation selectivity index)
            * peak_dff_sg (peak dF/F)
            * ptest_sg
            * time_to_peak_sg
            * duration_sg
        '''
        StaticGratings._log.info('Calculating peak response properties')

        peak = pd.DataFrame(index=range(self.numbercells), columns=('ori_sg','sf_sg', 'phase_sg', 'response_reliability_sg','osi_sg','peak_dff_sg','ptest_sg','time_to_peak_sg','duration_sg','cell_specimen_id'))
        cids = self.data_set.get_cell_specimen_ids()

        for nc in range(self.numbercells):
            cell_peak = np.where(self.response[:,1:,:,nc,0] == np.nanmax(self.response[:,1:,:,nc,0]))
            pref_ori = cell_peak[0][0]
            pref_sf = cell_peak[1][0]+1
            pref_phase = cell_peak[2][0]
            peak.cell_specimen_id.iloc[nc] = cids[nc]
            peak.ori_sg[nc] = pref_ori
            peak.sf_sg[nc] = pref_sf
            peak.phase_sg[nc] = pref_phase
            peak.response_reliability_sg[nc] = self.response[pref_ori, pref_sf, pref_phase, nc, 2]/0.48  #TODO: check number of trials
            pref = self.response[pref_ori, pref_sf, pref_phase, nc, 0]
            orth = self.response[np.mod(pref_ori+3, 6), pref_sf, pref_phase, nc, 0]
            peak.osi_sg[nc] = (pref-orth)/(pref+orth)
            peak.peak_dff_sg[nc] = pref
            groups = []
            
            for ori in self.orivals:
                for sf in self.sfvals[1:]:
                    for phase in self.phasevals:
                        groups.append(self.mean_sweep_response[(self.stim_table.spatial_frequency==sf)&(self.stim_table.orientation==ori)&(self.stim_table.phase==phase)][str(nc)])
            groups.append(self.mean_sweep_response[self.stim_table.spatial_frequency==0][str(nc)])

            _,p = st.f_oneway(*groups)
            peak.ptest_sg[nc] = p
            
            test_rows = (self.stim_table.orientation==self.orivals[pref_ori]) & \
                (self.stim_table.spatial_frequency==self.sfvals[pref_sf]) & \
                (self.stim_table.phase==self.phasevals[pref_phase])

            if len(test_rows) < 2:
                msg = "Static grating p value requires at least 2 trials at the preferred " 
                "orientation/spatial frequency/phase. Cell %d (%f, %f, %f) has %d." % \
                    (int(nc), self.orivals[pref_ori], self.sfvals[pref_sf], 
                     self.phasevals[pref_phase], len(test_rows))

                raise BrainObservatoryAnalysisException(msg)

            test = self.sweep_response[test_rows][str(nc)].mean()
            peak.time_to_peak_sg[nc] = (np.argmax(test) - self.interlength)/self.acquisition_rate
            test2 = np.where(test<(test.max()/2))[0]
            try:          
                peak.duration_sg[nc] = np.ediff1d(test2).max()/self.acquisition_rate
            except:
                pass

        return peak



import numpy as np
import allensdk.brain_observatory.roi_masks as roi_masks


def test_init_by_pixels():
    a = np.array([[0,0],[1,1],[1,0]])
    
    m = roi_masks.create_roi_mask(2, 2, [0,0,0,0], pix_list=a)
    
    mp = m.get_mask_plane()

    assert mp[0,0] == 1
    assert mp[1,1] == 1
    assert mp[1,0] == 0
    assert mp[1,1] == 1

    assert m.x == 0
    assert m.width == 2
    assert m.y == 0
    assert m.height == 2


def test_init_by_pixels_with_border():
    a = np.array([[1,1],[2,1]])
    
    m = roi_masks.create_roi_mask(3, 3, [1,1,1,1], pix_list=a)

    assert m.x == 1
    assert m.width == 2
    assert m.y == 1
    assert m.height == 1
    assert m.valid is False


def test_init_by_pixels_large():
    a = np.random.random((512,512))
    a[a > 0.5] = 1

    m = roi_masks.create_roi_mask(512, 512, [0,0,0,0], pix_list=np.argwhere(a))

    npx = len(np.where(a)[0])
    assert npx == len(np.where(m.get_mask_plane())[0])
